///////////////////////////////////////////////////////////////////////
//  Copyright (C) ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXCommonPersistence.cpp,v 1.1 2004/09/07 06:10:49 me133 Exp $
//
//    File name:	ILXCommonPersistence.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/4/99 5:47:04 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXCommonPersistence.cpp,v $
 * Revision 1.1  2004/09/07 06:10:49  me133
 * o Initial checkin to CVS
 * 
 * 
 * 43    4/10/02 8:20p Marce
 * o Fixed bug where we were still counting properties that should be
 * ignored.  This caused E_UNEXPECTED to be returned by Load() after
 * everything is loaded successfully.
 * 
 * 40    5/04/01 6:40p Marce
 * o I had moved the assert statement incorrectly
 * 
 * 39    4/30/01 3:53p Marce
 * o Added support for IgnorePersistedProperty.  Each control should store
 * the m_setDispidsToIgnore property, update it when
 * IgnorePersistedProperty is called, and pass it to
 * ILXCommonLoadFromStream() and ILXCommonSaveToStream().
 * 
 * 37    7/03/00 1:01p Marce
 * o Incremented version number since we changed the way we skip
 * properties when GetProperty() fails
 * o Modified Load() to skip VT_EMPTY properties.  VT_EMPTY is what we
 * save when GetProperty() fails.
 * o Now always writes a variant value during save even if we fail to
 * retrieve the property.
 * 
 * 36    6/26/00 6:07p Marce
 * o Incremented the version number
 * o RemoveSymbolFromStream only modifies streams with the new version
 * number so that it is compatible with old versions of controls.
 * o Added new argument to ILXCommonLoadFromStream so that we don't have
 * to use a global variable
 * 
 * 35    6/23/00 5:20p Marce
 * o Removed diagnostic output
 * 
 * 34    6/22/00 6:40p Marce
 * o Fixed a dormant bug where we would not actually ignore some unknown
 * properties properly.  This bug would manifest itself for BSTRs, Fonts,
 * or other non-primitive objects.
 * 
 * 33    6/22/00 6:32p Marce
 * o Fixed a bug that was preventing the symbol from being removed
 * 
 * 32    6/22/00 6:24p Marce
 * o We always have to read the entire variant from the stream because
 * some properties (e.g., Font) do not give correct values for their size
 * 
 * 31    6/22/00 6:18p Marce
 * o Fixed bug in diagnostics
 * 
 * 30    6/22/00 6:14p Marce
 * o Added diagnostics
 * 
 * 29    6/22/00 5:58p Marce
 * o Fixed bug in determining BSTR size during Save()
 * 
 * 28    6/22/00 5:29p Marce
 * o Fixed bug where the length of the BSTR data was not correct during
 * Save()
 * 
 * 27    6/22/00 5:21p Marce
 * o Fixed bug in Save() were the size of the BSTR was not saved correctly
 * 
 * 26    6/22/00 4:45p Marce
 * o Fixed bug where Symbol name was not cleared
 * 
 * 25    6/22/00 4:38p Marce
 * o Fixed bugs in RemoveSymbolFromStream()
 * 
 * 24    6/22/00 3:35p Marce
 * o Added RemoveSymbolFromStreamKludge()
 * 
 * 23    6/22/00 3:03p Marce
 * o Rolled-back to previous version by removing
 * ILXCommonLoadFromStream2().  No one should be calling this function
 * now.
 * 
 * 22    6/20/00 10:57p Marce
 * o You can now turn off property buffering by passing NULL for the apply
 * function
 * 
 * 21    6/19/00 7:24p Marce
 * o Changed typedef name
 * 
 * 20    6/19/00 6:58p Marce
 * o Now buffers data directory, instance name, and mode
 * 
 * 19    6/06/00 11:16a Marce
 * o Now allows for partial failure
 * 
 * 18    6/01/00 4:35p Marce
 * o Fixed a bug that happened when attempting to persist a property that
 * is an object (IUnknown or IDispatch).  If the property object does not
 * support IPersistStream, we would get an access violation.
 * 
 * 17    5/31/00 9:46p Marce
 * o Hex dispids are hard to understand
 * 
 * 16    5/31/00 9:44p Marce
 * o We were not seeing the trace messages for failed property loading
 * 
 * 15    5/26/00 9:37a Alexanderk
 * 
 * 14    4/28/00 6:08p Marce
 * o Added an optimization that works because we assume save() and load()
 * are called by the same version of the control (has the same property
 * map)
 * 
 * 13    4/28/00 5:55p Marce
 * o Inside the loop we need to use break instead of return
 * 
 * 12    4/28/00 5:37p Marce
 * o We don't need to use case insensitive string comparisons because we
 * control the saving and the loading
 * 
 * 11    4/28/00 5:36p Marce
 * o Previous kludge made for OptionMontage was removed.  Kludge never
 * worked in the first place and OM found a different solution.
 * 
 * 10    4/19/00 4:30p Marce
 * o Added a kludge for OptionMontage to allow it to return E_NOTIMPL to
 * indicate properties that are not valid for a given mode.  Since it is
 * rare that a property in the property map would return E_NOTIMPL for any
 * other reason, this kludge should be safe for all controls.
 * 
 * 9     2/11/00 12:51p Christ
 * 
 * 8     2/11/00 12:32p Chris
 * Update the version number for headline also expose version number
 * outside of this module.
 * 
 * 7     10/08/99 11:02a Rafaelc
 * The property count now contains only propeerty map entries with a
 * description.
 * 
 * 6     10/07/99 10:32a Rafaelc
 * o Now handles property entries that have no description (like
 * PROP_PAGE)
 * 
 * 5     10/05/99 1:54p Marce
 * o I am rolling back...
 * 
 * 4     10/05/99 1:51p Marce
 * o Now ignores monitor persistence format
 * 
 * 3     10/04/99 6:13p Marce
 * o Comments
 * 
 * 2     10/04/99 5:59p Marce
 */

#include "ILXCommonPersistence.h"	// Me

#include "ILXCommon.h"				// DISPID_ILX_*

extern CComModule _Module;
#include <atlcom.h>

// Constants

const DWORD nILX_COMMON_PERSIST_BUGGY_VER = 0x0702;

//*********************************************************************
// FUNCTION		:	ILXCommonSaveToStream
//					
// DESCRIPTION	:	Save properties to the stream specifying the
//					property name, dispid, and raw data.  The stream
//					is therefore self-describing.
//					
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   pStm	
//   pMap	
//   pThis	
//   pUnk	
//					
// COMMENTS		:	During load() we can skip properties we don't
//					recognize by checking the property name.
//					
//
//	     HEADER                             PROPERTY RECORD
//                       
//     4       4          4        NameSize       1      4       8     DataSize
// +-------+-------+  +--------+--------------+------+------+--------+---------+
// |  Ver  | Count |  |NameSize|     Name     |IsRaw?|DispId|DataSize|  Data   |
// +-------+-------+  +--------+--------------+------+------+--------+---------+
//
//
//*********************************************************************
HRESULT ILXCommonSaveToStream(	IStream* pStm, 
								ATL_PROPMAP_ENTRY* pMap,
								void* pThis,
								IUnknown* pUnk,
								const TypeSetDispids* psetDispidsToIgnore)
{
	_ASSERTE(pMap);
	_ASSERTE(pUnk);
	_ASSERTE(pStm);
	_ASSERTE(pThis);

	HRESULT hr = S_OK;
	HRESULT hrRet = S_OK;

	// Cache the end of the set
	TypeSetDispids::const_iterator itDispids_end = 0;
	if (psetDispidsToIgnore)
	{
		itDispids_end = psetDispidsToIgnore->end();
	}

	// Step 1:	Write out the version number

	hr = pStm->Write(&nILX_COMMON_PERSIST_VER, sizeof(DWORD), NULL);
	if (FAILED(hr)) return hr;

	// Step 2:	Write out the number of properties

	long nPropCount = 0;
	for (int j = 0; pMap[j].pclsidPropPage; ++j)
	{
		// Only count the property if the description is non-NULL and its
		// not a property we've been told to ignore
		if (pMap[j].szDesc &&
			(!psetDispidsToIgnore ||
			psetDispidsToIgnore->find(pMap[j].dispid) == itDispids_end))
		{
			nPropCount++;
		}
	}

	hr = pStm->Write(&nPropCount, sizeof(long), NULL);
	if (FAILED(hr)) return hr;

	// Step 3:	Write out all the properties

	CComPtr<IDispatch> spiDispatch;
	const IID* piidOld = NULL;
	for (int i = 0; pMap[i].pclsidPropPage; ++i)
	{
		ULARGE_INTEGER cbWrite;
		cbWrite.QuadPart = 0;

		if (!pMap[i].szDesc)
		{
			continue;
		}

		if (psetDispidsToIgnore && 
			psetDispidsToIgnore->find(pMap[i].dispid) != itDispids_end)
		{
			// We've been told to not save this property
			continue;
		}

		// Step 4:	Write out the length of the name
		long nNameSize = sizeof(TCHAR) * _tcslen(pMap[i].szDesc);
		hr = pStm->Write(	&nNameSize, 
							sizeof(nNameSize), 
							NULL);
		if (FAILED(hr)) return hr;

		// Step 5:	Write out the name of the property
		hr = pStm->Write(	pMap[i].szDesc, 
							nNameSize, 
							NULL);
		if (FAILED(hr)) return hr;

		bool bIsRaw = pMap[i].dwSizeData != 0;

		// Step 6:	Say if the property is a raw value
		hr = pStm->Write(	&bIsRaw, 
							sizeof(bool), 
							NULL);
		if (FAILED(hr)) return hr;

		// Step 7:	Give the property's dispid (may be 0)
		hr = pStm->Write(	&pMap[i].dispid, 
							sizeof(DISPID), 
							NULL);
		if (FAILED(hr)) return hr;

		// check if raw data entry
		if (pMap[i].dwSizeData != 0)
		{
			cbWrite.QuadPart = pMap[i].dwSizeData;

			// Step 8:	Write out the size of the data
			hr = pStm->Write(&cbWrite, sizeof(cbWrite), NULL);
			if (FAILED(hr)) return hr;

			// Step 9:	Write out the actual data
			void* pData = (void*) (pMap[i].dwOffsetData + (DWORD)pThis);
			hr = pStm->Write(pData, pMap[i].dwSizeData, NULL);
			if (FAILED(hr)) return hr;
			continue;
		}

		CComVariant var; // Initialized to VT_EMPTY

		// Step 10:	Get the IDispatch ptr
		if (pMap[i].piidDispatch != piidOld)
		{
			spiDispatch.Release();
			hr = pUnk->QueryInterface(*pMap[i].piidDispatch, (void**) &spiDispatch);
			if (FAILED(hr))
			{
				AtlTrace(_T("Failed to get a dispatch pointer for property %s (DISPID %d)\n"), 
					pMap[i].szDesc, pMap[i].dispid);

				_ASSERTE(var.vt == VT_EMPTY);

				// We failed to save this property but try to save the rest
				if (hrRet == S_OK)
				{
					hrRet = hr;
				}
			}
			else
			{
				piidOld = pMap[i].piidDispatch;
			}
		}

		// Step 11:	Get the property value
		if (SUCCEEDED(hr) && spiDispatch)
		{
			hr = CComDispatchDriver::GetProperty(spiDispatch, pMap[i].dispid, &var);
			if (FAILED(hr))
			{
				AtlTrace(_T("GetProperty failed on %s (DISPID %d)\n"), 
					pMap[i].szDesc, pMap[i].dispid);

				_ASSERTE(var.vt == VT_EMPTY);

				// We failed to save this property but try to save the rest
				if (hrRet == S_OK)
				{
					hrRet = hr;
				}
			}
		}

		// Determine the size of the data to write

		switch (var.vt)
		{
		case VT_UNKNOWN:
		case VT_DISPATCH:
			{
				CComPtr<IPersistStream> spiPersistStream;
				if (var.punkVal != NULL)
				{
					hr = var.punkVal->QueryInterface(IID_IPersistStream, 
													(void**) &spiPersistStream);
					if (FAILED(hr)) return hr;
				}

				if (spiPersistStream)
				{
					hr = spiPersistStream->GetSizeMax(&cbWrite);
					if (FAILED(hr)) return hr;
				}
				else
				{
					// CLSID_NULL is what will be written to the stream
					// since this object does not support IPersistStream
					cbWrite.QuadPart = sizeof(CLSID);
				}

				break;
			}

		case VT_UI1:
		case VT_I1:
			cbWrite.QuadPart = sizeof(BYTE);
			break;
		case VT_I2:
		case VT_UI2:
		case VT_BOOL:
			cbWrite.QuadPart = sizeof(short);
			break;
		case VT_I4:
		case VT_UI4:
		case VT_R4:
		case VT_INT:
		case VT_UINT:
		case VT_ERROR:
			cbWrite.QuadPart = sizeof(long);
			break;
		case VT_R8:
		case VT_CY:
		case VT_DATE:
			cbWrite.QuadPart = sizeof(double);
			break;
		case VT_BSTR:

			// The length of the BSTR is always written first
			cbWrite.QuadPart = sizeof(ULONG);

			// If the BSTR is non-NULL then the entire BSTR
			// plus terminating NULL is written out
			if (var.bstrVal)
			{
				cbWrite.QuadPart +=	::SysStringByteLen(var.bstrVal) + 
									sizeof(OLECHAR);
			}

			break;
		case VT_EMPTY:
			cbWrite.QuadPart = 0;
			break;
		default:
			break;
		}

		// var.WriteToStream() adds on the vartype
		cbWrite.QuadPart += sizeof(VARTYPE);

		// Step 12:	Write out the size of the data
		hr = pStm->Write(&cbWrite, sizeof(cbWrite), NULL);
		if (FAILED(hr)) return hr;

		// Step 13:	Write out the actual data
		hr = var.WriteToStream(pStm);
		if (FAILED(hr)) return hr;
	}

	return hrRet;
}

//*********************************************************************
// FUNCTION		:	ILXCommonLoadFromStream
//					
// DESCRIPTION	:	Load properties from a self-describing or binary
//					stream.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   pStm	
//   pMap	
//   pThis	
//   pUnk	
//					
// COMMENTS		:	We support the IlxCommon format (see
//					ILXCommonSaveToStream) and the Atl 3.0 binary
//					format (AtlIPersistStreamInit_Load).
//					
//*********************************************************************
HRESULT ILXCommonLoadFromStream(IStream* pStm, 
								ATL_PROPMAP_ENTRY* pMap, 
								void* pThis, 
								IUnknown* pUnk,
								DWORD* pdwPersistenceFormat,
								const TypeSetDispids* psetDispidsToIgnore)
{
	_ASSERTE(pMap);
	_ASSERTE(pUnk);
	_ASSERTE(pStm);
	_ASSERTE(pThis);
	
	HRESULT hr = S_OK;
	HRESULT hrRet = S_OK;

	// Cache the end of the set
	TypeSetDispids::const_iterator itDispids_end = 0;
	if (psetDispidsToIgnore)
	{
		itDispids_end = psetDispidsToIgnore->end();
	}

	// Step 1:	Read the persistence format version

	DWORD dwVer = 0;
	hr = pStm->Read(&dwVer, sizeof(DWORD), NULL);
	if (FAILED(hr)) return hr;

	if (pdwPersistenceFormat)
	{
		*pdwPersistenceFormat = dwVer;
	}

	// Ignore the Monitor format
	if (dwVer == 7100) // Got version number from ILXMonitor.h
	{
		return S_OK;
	}
	else if (dwVer < nILX_COMMON_PERSIST_BEGIN_VER) 
	{
		// We are dealing with an older persistence version.  We
		// assume that the older version is the default ATL 3.0
		// AtlIPersistStreamInit_Save() version so we can try to
		// use AtlIPersistStreamInit_Load() to load it.

		// First we have to reset the stream pointer to the beginning
		// of the stream
		LARGE_INTEGER nOffset;
		nOffset.QuadPart = sizeof(DWORD);
		nOffset.QuadPart *= -1;
		hr = pStm->Seek(nOffset, STREAM_SEEK_CUR, 0);
		if (FAILED(hr)) return hr;

		return AtlIPersistStreamInit_Load(pStm, pMap, pThis, pUnk);
	}

	// Step 2:	Read the total number of properties in the stream

	long nTotalProperties = 0;
	hr = pStm->Read(&nTotalProperties, sizeof(nTotalProperties), NULL);
	if (FAILED(hr)) return hr;

	if (nTotalProperties < 0) return E_UNEXPECTED;

	// Step 3:	Count the properties in the class's property map

	long nPropMapCount = 0;
	for(; pMap[nPropMapCount].pclsidPropPage; ++nPropMapCount) ; // Empty loop

	// Step 4:	Read all the properties

	CComPtr<IDispatch> spiDispatch;
	const IID* piidOld = NULL;

	for(int j = 0; j < nTotalProperties; ++j)
	{
		// Step 5:	Read the property name length
		long nNameSize = 0;
		hr = pStm->Read(&nNameSize, sizeof(nNameSize), NULL);
		if (FAILED(hr) || nNameSize == 0) 
		{
			hrRet = FAILED(hr) ? hr : E_UNEXPECTED;
			break;	// Failure!
		}

		// Step 6:	Read the property name

		long nNameLen = nNameSize / sizeof(TCHAR);
		TCHAR* szName = new TCHAR[nNameLen + 1];

		hr = pStm->Read(szName, nNameSize, NULL);
		if (FAILED(hr)) break; // Failure!
		szName[nNameLen] = _T('\0');

		bool bIsRaw = false;

		// Step 7:	See if the property is a raw value
		hr = pStm->Read(&bIsRaw, sizeof(bool), NULL);
		if (FAILED(hr)) 
		{
			hrRet = hr;
			break;	// Failure!
		}

		// Step 8:	Get the dispid of the property (may be null)
		DISPID dispid = 0;
		hr = pStm->Read(&dispid, sizeof(DISPID), NULL);
		if (FAILED(hr)) 
		{
			hrRet = hr;
			break;	// Failure!
		}

		ULARGE_INTEGER cbRead;
		cbRead.QuadPart = 0;

		// Step 9:	Read the size of the data
		hr = pStm->Read(&cbRead, sizeof(cbRead), NULL);
		if (FAILED(hr)) 
		{
			hrRet = hr;
			break;	// Failure!
		}

		// Step 10:	Find the property in the property map

		int i = 0;

		// Optimization: Assume that the property map has not changed since
		// we last saved.
		if (j >= 0 &&
			j < nPropMapCount &&
			pMap[j].szDesc &&
			_tcscmp(pMap[j].szDesc, szName) == 0)
		{
			i = j;
		}
		else
		{
			// Find the raw property entry in the property map
			for (i = 0; pMap[i].pclsidPropPage; ++i)
			{ 
				if (pMap[i].szDesc &&
					_tcscmp(pMap[i].szDesc, szName) == 0)
				{
					break;
				}
			}
		}

		// We don't need the name anymore
		delete[] szName;
		szName = 0;

		// Step 11:	Read the actual data

		// Read raw data value
		if (bIsRaw)
		{
			if (!pMap[i].pclsidPropPage)
			{
				// We don't know this raw property so skip it
				LARGE_INTEGER cbSeek;
				cbSeek.QuadPart = cbRead.QuadPart;

				hr = pStm->Seek(cbSeek, STREAM_SEEK_CUR, 0);
			}
			else
			{
				_ASSERTE(cbRead.QuadPart == pMap[i].dwSizeData);

				// Read the data directly into the raw data member
				void* pData = (void*) (pMap[i].dwOffsetData + (DWORD)pThis);
				hr = pStm->Read(pData, pMap[i].dwSizeData, NULL);
			}

			if (FAILED(hr)) 
			{
				// We can't necessarily recover from this error since the
				// stream may not have been advanced correctly
				hrRet = hr;
				break;
			}

			continue;
		}

		// Else, Read the VARIANT data value

		CComVariant var;

		hr = var.ReadFromStream(pStm);
		if (FAILED(hr) && hrRet == S_OK)
		{
			// We can't necessarily recover from this error since the
			// stream may not have been advanced correctly
			hrRet = hr;
			break;
		}

		// Step 12:	Skip this property?
		//
		//			This is where properties should be skipped if necessary
		//			since the stream has been advanced correctly.

		if (!pMap[i].pclsidPropPage)
		{
			// We don't know this property so skip it
			continue;
		}

		_ASSERTE(memcmp(&dispid, &pMap[i].dispid, sizeof(DISPID)) == 0);

		if (var.vt == VT_EMPTY)
		{
			// This property was not saved properly during Save() so skip it.
			continue;
		}

		if (psetDispidsToIgnore && 
			psetDispidsToIgnore->find(pMap[i].dispid) != itDispids_end)
		{
			// We've been told to ignore this property
			continue;
		}

    // Step 13:	Get the IDispatch interface

		if (pMap[i].piidDispatch != piidOld)
		{
			spiDispatch.Release();
			hr = pUnk->QueryInterface(*pMap[i].piidDispatch, (void**)&spiDispatch);
			if (FAILED(hr))
			{
				AtlTrace(_T("Failed to get a dispatch pointer for property %s (DISPID %d)\n"), 
					pMap[i].szDesc, pMap[i].dispid);

				// Remember the error code and continue processing the rest
				// of the properties
				if (hrRet == S_OK)
				{
					hrRet = hr;
				}
				
				continue;
			}

			piidOld = pMap[i].piidDispatch;
		}

		// Step 14:	Call the appropriate put_ method

		hr = CComDispatchDriver::PutProperty(spiDispatch, pMap[i].dispid, &var);
		if (FAILED(hr))
		{
			AtlTrace(_T("PutProperty failed on %s (DISPID %d)\n"), 
				pMap[i].szDesc, pMap[i].dispid);

			// Remember the error code and continue processing the rest
			// of the properties
			if (hrRet == S_OK)
			{
				hrRet = hr;
			}
		}
	}

	return hrRet;
}

HRESULT RemoveSymbolFromStreamKludge(IStream* piStreamOriginal)
{
	HRESULT hr = S_OK;

	_ASSERTE(piStreamOriginal);
	if (!piStreamOriginal) return E_POINTER;
	
	CComPtr<IStream> spiStreamClone;

	hr = piStreamOriginal->Clone(&spiStreamClone);
	if (FAILED(hr)) return hr;
	if (!spiStreamClone) return E_UNEXPECTED;

	// Step 1:	Do nothing if we don't recognize the stream format

	DWORD dwVer = 0;
	hr = spiStreamClone->Read(&dwVer, sizeof(DWORD), NULL);
	if (FAILED(hr)) return hr;

	// Ignore the Monitor format
	if (dwVer == 7100) // Got version number from ILXMonitor.h
	{
		return S_OK;
	}
	else if (dwVer < nILX_COMMON_PERSIST_BEGIN_VER) 
	{
		// We are dealing with an older persistence version.  We
		// assume that the older version is the default ATL 3.0
		// AtlIPersistStreamInit_Save() version so we should not
		// try to modify it.

		return S_OK;
	}
	else if (dwVer == nILX_COMMON_PERSIST_BUGGY_VER)
	{
		// This version of the stream did not know how to skip
		// properties correctly so we cannot modify it.
		return S_OK;
	}

	// Step 2:	Read the total number of properties in the stream

	long nTotalProperties = 0;
	hr = spiStreamClone->Read(&nTotalProperties, sizeof(nTotalProperties), NULL);
	if (FAILED(hr)) return hr;

	if (nTotalProperties < 0) return S_FALSE;

	// Step 3:	Find the Symbol property

	LARGE_INTEGER cbOffset;
	ULARGE_INTEGER cbRead;

	for(int j = 0; j < nTotalProperties; ++j)
	{
		// Step 4:	Read the property name length
		long nNameSize = 0;
		hr = spiStreamClone->Read(&nNameSize, sizeof(nNameSize), NULL);
		if (FAILED(hr)) return hr;
		if (nNameSize == 0) return E_UNEXPECTED;

		// Step 5:	Skip past the property name
		cbOffset.QuadPart = nNameSize;
		hr = spiStreamClone->Seek(cbOffset, STREAM_SEEK_CUR, 0);
		if (FAILED(hr)) return hr;

		bool bIsRaw = false;

		// Step 6:	See if the property is a raw value
		hr = spiStreamClone->Read(&bIsRaw, sizeof(bool), NULL);
		if (FAILED(hr)) return hr;

		// Step 7:	Get the dispid of the property (may be null)
		DISPID dispid = DISPID_UNKNOWN;
		hr = spiStreamClone->Read(&dispid, sizeof(DISPID), NULL);
		if (FAILED(hr)) return hr;

		// Step 8:	If the property is the symbol property then
		//			change the dispid to DISPID_UNKNOWN so that
		//			the controls will ignore the property when
		//			Load() is called.
		if (dispid == DISPID_ILX_SYMBOL)
		{
			// Step 8a:	Backtrack the stream to point to the DISPID
			cbOffset.QuadPart = sizeof(DISPID);
			cbOffset.QuadPart *= -1;
			hr = spiStreamClone->Seek(cbOffset, STREAM_SEEK_CUR, 0);
			if (FAILED(hr)) return hr;

			// Step 8b:	Change the dispid to DISPID_UNKNOWN
			dispid = DISPID_UNKNOWN;
			hr = spiStreamClone->Write(&dispid, sizeof(DISPID), NULL);
			if (FAILED(hr)) return hr;

			// Step 8c:	Backtrack the stream to point to the property name
			cbOffset.QuadPart = sizeof(DISPID) + sizeof(bool) + nNameSize;
			cbOffset.QuadPart *= -1;
			hr = spiStreamClone->Seek(cbOffset, STREAM_SEEK_CUR, 0);
			if (FAILED(hr)) return hr;

			// Step 8d:	Clear the Symbol name
            TCHAR cZero = _T('\0');
			hr = spiStreamClone->Write(&cZero, sizeof(cZero), NULL);
			if (FAILED(hr)) return hr;

			// We are done!
			return S_OK;
		}

		// Step 9:	Skip to the next property

	    // Read the size of the data
	    cbRead.QuadPart = 0;
	    hr = spiStreamClone->Read(&cbRead, sizeof(cbRead), NULL);
	    if (FAILED(hr)) return hr;

		if (bIsRaw)
		{
			cbOffset.QuadPart = cbRead.QuadPart;
			hr = spiStreamClone->Seek(cbOffset, STREAM_SEEK_CUR, 0);
			if (FAILED(hr)) return hr;
		}
		else
		{
		    // Skip past the data
			CComVariant var;
			hr = var.ReadFromStream(spiStreamClone);
			if (FAILED(hr)) return hr;
		}
    }

	return S_FALSE; // We did not find a Symbol property
}
