///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/IlxContextImpl.h,v 1.1 2004/09/07 06:10:50 me133 Exp $
//
//    File name:	IlxContextImpl.h
//   Created By:	Marc Eaddy
//   Created On:	04/09/2002 1:49:51 PM
//
//      Purpose:	
//
//  Description:	XML is case-sensitive!
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: IlxContextImpl.h,v $
 * Revision 1.1  2004/09/07 06:10:50  me133
 * o Initial checkin to CVS
 * 
 * 
 * 23    8/30/02 8:24p Marce
 * o Now supports arbitrary transferable property transfer via RMB
 * 
 * 22    8/09/02 10:39a Jason
 * Added bstrIS_FONT_ATTR.
 * 
 * 21    7/31/02 2:44p Marce
 * o Added GetReasonString()
 * 
 * 19    6/07/02 6:11p Marce
 * o Now keeps the Drill Down color properties in-sync with the Context
 * properties
 * 
 * 18    6/04/02 8:21p Marce
 * o Now CContextCallback also keeps track of which context properties
 * were modified by SetContextToXml()
 * o Moved implementation of CContextCallback into cpp file
 * 
 * 17    5/31/02 1:03p Marce
 * o Added HasContextProperties()
 * 
 * 16    5/16/02 6:17p Marce
 * o If you want to find out if a property exists you don't have to
 * specify a result variant in GetContext2.  It returns S_OK if the
 * property exists and S_FALSE otherwise.
 * 
 * 15    5/16/02 5:36p Marce
 * o Fixed bug where typing in a symbol in News caused a double navigate
 * 
 * 14    5/14/02 3:08p Marce
 * o Default delimiter changed to semicolon to match the default delimiter
 * used by IIlxProperties
 * 
 * 13    5/03/02 11:17a Marce
 * o BeforeSetContext was somehow messing up the heap because it was
 * calling VariantChangeType() on the variant passed in via the
 * marshaller.  It was causing a variant that was five levels up the call
 * stack to change.
 * o Lesson learned: Don't call VariantChangeType() directly on a variant
 * that's been passed to you by scripting or the marshaller!
 * 
 * 12    5/02/02 12:37p Marce
 * o Made some functions public
 * 
 * 11    4/30/02 9:48p Marce
 * o Added support for UrlQuery so that all values are url canonicalized
 * 
 * 10    4/30/02 8:43p Marce
 * o Implemented saving context to a file path
 * o Now longer need to expose LoadContext() and SaveContext()
 * 
 * 9     4/27/02 2:06a Marce
 * o Created a callback interface to allow C++ clients (controls) to be
 * notified before and after a context property changes
 * 
 * 8     4/24/02 10:13p Marce
 * o Added support for NoPersist
 * o Added GetContext2Impl() and SetContext2Impl()
 * o Fixed bug when removing the first property
 * 
 * 7     4/24/02 6:19p Marce
 * o Added preliminary code for supporting saving/load XML nodes
 * o Created more generic RemoveContextProperty()
 * 
 * 6     4/23/02 4:33p Marce
 * o Minor
 * 
 * 5     4/23/02 4:04p Marce
 * o Changed some constant and enumeration names
 * o Separated Action from Reason in CContextOptions
 * 
 * 4     4/17/02 12:21p Marce
 * o SetContext2 now requires the varOptions parameter
 * o Added strings for all the possible Reasons
 * 
 * 3     4/16/02 9:08p Marce
 * o Now implements most of the new Context API
 * o Added comments
 * 
 * 2     4/11/02 5:58p Marce
 * o No longer need to fire event from IlxContextImpl.h.  Now we can share
 * this file with ActiveILX.
 * 
 * 1     4/11/02 12:24p Marce
 */

#ifndef INCLUDED_ILX_CONTEXT_IMPL_H
#define INCLUDED_ILX_CONTEXT_IMPL_H

#ifndef INCLUDED_MSXML_TYPE_LIBRARY
#define INCLUDED_MSXML_TYPE_LIBRARY
#import "msxml.dll" rename_namespace("MSXML"), named_guids, raw_interfaces_only, raw_native_types
#endif // INCLUDED_MSXML_TYPE_LIBRARY

#include "IlxLogHelper.h"			// Logging macros
#include "XmlLogHelper.h"			// LogXmlError()

#include <xstring>					// STL string

// Type defines
typedef std::basic_string<TCHAR> tstring;

// Constants

// Context is the top-level XML document element ("<Context/>")
static CComBSTR bstrCONTEXT_ELEMENT					= L"Context";

// Prop is the tag name for the XML property element ("<Context><Prop/></Context>")
static CComBSTR bstrPROP_ELEMENT					= L"Prop";

// Attribute names that can be used for context property elements

static CComBSTR bstrNAME_ATTR						= L"Name";
static CComBSTR bstrVALUE_ATTR						= L"Value";
static CComBSTR bstrDISPLAY_ATTR					= L"Display";
static CComBSTR bstrSUBMENU_ATTR					= L"SubMenu";
static CComBSTR bstrDEFAULT_ATTR					= L"Default";
static CComBSTR bstrNO_PERSIST_ATTR					= L"NoPersist";
static CComBSTR bstrIS_COLOR_ATTR					= L"IsColor";
static CComBSTR bstrIS_FONT_ATTR					= L"IsFont";
static CComBSTR bstrIS_TRANSFERABLE_ATTR			= L"IsTrans";

// Recognized property names (<Prop Name="Symbol" Value="IBM"/>)
static CComBSTR bstrSYMBOL_PROP_NAME				= L"Symbol";
static CComBSTR bstrDRILL_DOWN_COLOR_PROP_NAME		= L"DrillDownColor";
static CComBSTR bstrDRILL_DOWN_HOVER_COLOR_PROP_NAME= L"DrillDownHoverColor";

// Recognized options and values in the options string
// E.g. "Reason=discovery"

static const wchar_t* szOPTION_NAME_REASON			= L"Reason";
static const wchar_t* szREASON_DRAG_AND_DROP			= L"dnd";
static const wchar_t* szREASON_RMB						= L"rmb";
static const wchar_t* szREASON_PROGRAMMATIC				= L"prog";
static const wchar_t* szREASON_UI						= L"ui";
static const wchar_t* szREASON_DISCOVERY				= L"discovery";
static const wchar_t* szREASON_SHORTCUT					= L"shortcut";

static const wchar_t* szOPTION_NAME_ACTION			= L"Action";
static const wchar_t* szACTION_QUERY					= L"query";
static const wchar_t* szACTION_URL_QUERY				= L"urlquery";
static const wchar_t* szACTION_REMOVE					= L"remove";
static const wchar_t* szACTION_REMOVE_ATTRIBUTE			= L"removeattr";
static const wchar_t* szACTION_LOAD						= L"load";
static const wchar_t* szACTION_SAVE						= L"save";
static const wchar_t* szACTION_REPLACE					= L"replace";

static const wchar_t* szOPTION_NAME_FORMAT			= L"Format";
static const wchar_t* szFORMAT_DOCUMENT					= L"document";
static const wchar_t* szFORMAT_NODE						= L"node";
static const wchar_t* szFORMAT_STRING					= L"string";
static const wchar_t* szFORMAT_VARIANT					= L"variant";

static const wchar_t* szOPTION_NAME_ATTRIBUTE		= L"Attribute";
static const wchar_t* szOPTION_NAME_DELIMITER		= L"Delimiter";
static const wchar_t* szOPTION_NAME_ERASE_UNKNOWNS	= L"EraseUnknowns";
static const wchar_t* szOPTION_NAME_UPDATE_ONLY		= L"UpdateOnly";

// Options must be separated by a semi-colon
// (ex "Reason=discovery;Format=string")
static const wchar_t cOPTIONS_DELIMITER				= L';';

// The default property delimiter (see CContextOptions::cDelimiter below)
static const wchar_t cDEFAULT_PROPERTY_DELIMITER	= L';';
static const wchar_t cDEFAULT_QUERY_PROPERTY_DELIMITER = L'%';

// Web properties should use '$'.  If the property isn't found and isn't
// removed, '$' in the URL won't hurt anything (as opposed to '&').
static const wchar_t cWEB_PROPERTY_DELIMITER		= L'$';

// Shortcuts
static CComBSTR bstrOPTION_NAME_FORMAT				= szOPTION_NAME_FORMAT;
static CComBSTR bstrFORMAT_STRING					= szFORMAT_STRING;
static CComBSTR bstrOPTION_NAME_ATTRIBUTE			= szOPTION_NAME_ATTRIBUTE;
static CComVariant varSYMBOL_PROP_NAME				= bstrSYMBOL_PROP_NAME;
static CComVariant varDRILL_DOWN_COLOR_PROP_NAME	= bstrDRILL_DOWN_COLOR_PROP_NAME;
static CComVariant varDRILL_DOWN_HOVER_COLOR_PROP_NAME	= bstrDRILL_DOWN_HOVER_COLOR_PROP_NAME;
static CComVariant varEMPTY;

// Enumerations

enum EContextReason
{
	CTXT_REASON_UNINITIALIZED	= -1,

	// These four reasons are binary compatible with the
	// ILX_PROPERTY_CHANGE_SOURCE enumeration

	CTXT_REASON_DRAG_AND_DROP	= 0,	// ILX_DRAG_AND_DROP
	CTXT_REASON_RMB				= 1,	// ILX_RMB
	CTXT_REASON_PROGRAMMATIC	= 2,	// ILX_PROGRAMMATIC
	CTXT_REASON_UI				= 3,	// ILX_UI

	CTXT_REASON_DISCOVERY		= 1000,
	CTXT_REASON_SHORTCUT		= 1001,
};

enum EContextAction
{
	CTXT_ACTION_UNINITIALIZED	= -1,

	CTXT_ACTION_QUERY			= 2001,
	CTXT_ACTION_URL_QUERY		= 2002,
	CTXT_ACTION_REMOVE			= 2003,
	CTXT_ACTION_REMOVE_ATTRIBUTE= 2004,
	CTXT_ACTION_REPLACE			= 2005,
	CTXT_ACTION_LOAD			= 2006,
	CTXT_ACTION_SAVE			= 2007,
};

enum EContextFormat
{
	CTXT_FORMAT_UNINITIALIZED	= -1,

	CTXT_FORMAT_DOCUMENT		= 3001,
	CTXT_FORMAT_NODE			= 3002,
	CTXT_FORMAT_STRING			= 3003,
	CTXT_FORMAT_VARIANT			= 3004,
};

// Forward declarations
interface IILXProperties;

//*****************************
// CContextOptions
//*****************************
//
// This class holds GetContext2 and SetContext2 options that have been
// parsed from the varOptions VARIANT parameter.
//
class CContextOptions
{
public:

	// CTORS

	CContextOptions();	// Default CTOR
	
	CContextOptions(VARIANT varOptions);
		// Parses the variant to obtain the comma-delimited list of
		// options and option values.  If an option name or value
		// isn't recognized it is ignored.  Parsing errors are logged.

	CContextOptions(const EContextReason& eReason);

	CContextOptions(const EContextAction& eAction);

	CContextOptions(const CContextOptions&);	// COPY CTOR

	CContextOptions& operator=(const CContextOptions&); // ASSIGNMENT

	~CContextOptions(); // DTOR

	HRESULT init(VARIANT varOptions);

	void setDelimiter(const TCHAR cDelimiter);
	const TCHAR getDelimiter() const;

	bool isDelimiterSet() const;

public: // DATA

	EContextReason eReason;

	EContextAction eAction;

	EContextFormat eFormat;

	CComBSTR bstrAttributeName;
		// The name of the attribute to get or set.  Default is "Value".

	bool bEraseUnknowns;
		// Only applies when eReason=ILX_CTXT_QUERY. Whether or not to
		// remove properties from the output string when they aren't found.
		//
		// E.g. If Name="Joe" and query string is "My name is $Joe$
		//

	bool bUpdateOnly;
		// Applications can specify an AfterSetContext callback function
		// which is called after a property is modified.  If bUpdateOnly
		// is set to true then AfterSetContext will not be called.

private: // DATA

	bool bDelimiterIsSet;

	TCHAR cDelimiter;
		// The property delimiter used when specifying a list of property names
		// (e.g. "Symbol,Code,AccountNumber"), when specifying a context query string
		// ("http://gendb.ilx.net?joe=$JOE$"), and when formatting context as an
		// output string ("Symbol=IBM,Code=123")
};

const wchar_t* GetReasonString(const CContextOptions& options);

//*****************************
// CContextCallback
//*****************************
class CContextCallback
{
public:

	CContextCallback();
	~CContextCallback();

	HRESULT StartTrackingContextChanges();

	HRESULT ClearContextChanges();
	
	HRESULT AddContextChange(	/*[in]*/ MSXML::IXMLDOMElement* piPropElement,
								/*[in]*/ const CContextOptions& options);

	virtual HRESULT BeforeSetContext(	/*[in]*/ MSXML::IXMLDOMElement* piPropElement,
										/*[in]*/ const CContextOptions& options,
										/*[in]*/ BSTR bstrPropertyName,
										/*[in]*/ BSTR bstrAttributeName,
										/*[in]*/ VARIANT varOldValue,
										/*[in,out]*/ VARIANT varNewValue);

	virtual HRESULT AfterSetContext(	/*[in]*/ MSXML::IXMLDOMElement* piPropElement,
										/*[in]*/ const CContextOptions& options,
										/*[in]*/ BSTR bstrPropertyName,
										/*[in]*/ BSTR bstrAttributeName,
										/*[in]*/ VARIANT varOldValue,
										/*[in]*/ VARIANT varNewValue);

private: // NOT IMPLEMENTED
	CContextCallback(const CContextCallback&); // COPY CTOR
	CContextCallback& operator=(const CContextCallback&); // ASSIGNMENT

protected: // DATA
	
	CComPtr<MSXML::IXMLDOMDocument>	m_spiXmlContextChanges;
};

//*****************************
// CIlxContextImpl
//*****************************
//
// 1. Control must derive from CComControlBase<T>, CComControl<T>
//
// 2. Must have the following entries in the PROP_MAP
//
//	PROP_ENTRY("InstanceName", DISPID_ILX_INSTANCE_NAME, CLSID_NULL)
//	PROP_ENTRY("DataDirectory", DISPID_ILX_DATA_DIRECTORY, CLSID_NULL)
//
// 3. Should have the following macros specified in the IDL file
//
//	MTHD_GET_CONTEXT2
//	MTHD_SET_CONTEXT2
//	ON_CONTEXT_CHANGED
//
// 4. Derive from: (substitute ILXBrowser for your class, iface, iid, and libid)
//
//    If you currently derive form CStockPropImpl<> use
//
// public CIlxContextImpl<CILXBrowser, CStockPropImpl<CILXBrowser, IILXBrowser,  &IID_IILXBrowser, &LIBID_ILXBROWSERLib> >
//
//    or, if you currently derive from CDispatchImpl<> use
//
// public CIlxContextImpl<CILXBrowser, CDispatchImpl<IILXBrowser,  &IID_IILXBrowser, &LIBID_ILXBROWSERLib> >
//
//    otherwise use
//
// public CIlxContextImpl<CILXBrowser>
//
template <class T, class Base=T>
class ATL_NO_VTABLE CIlxContextImpl : public Base, public CContextCallback
{
public:

	typedef CIlxContextImpl<T, Base> TypeMyContextImpl;

	CIlxContextImpl()
		: m_spiXmlContext(0)
	{ }
	
	virtual ~CIlxContextImpl()
	{ }

	HRESULT FinalConstruct(bool bTrackContextChanges)
	{
		HRESULT hr = S_OK;

		hr = CreateContextDocument(&m_spiXmlContext, true);
		VERIFY_HR(hr, _T("Failed to create Context XML document."), ILXLOG_ERROR);

		if (bTrackContextChanges)
		{
			hr = StartTrackingContextChanges();
			VERIFY_HR(hr, _T("Failed to track context changes."), ILXLOG_ERROR);
		}

		return hr;
	}

	// You must call FinalRelease() in the control's FinalRelease()

	void FinalRelease()
	{ }

	// *******************************************************************
	// Context Properties
	// *******************************************************************

	//
	// GetContext2() is called to retrieve one or more context properties
	//
    STDMETHODIMP GetContext2(	/*[in,optional]*/ VARIANT varContext,
								/*[in,optional]*/ VARIANT varOptions,
								/*[out,retval]*/ VARIANT* pvarResult)
	{
		return GetContext2Impl(	varContext, 
								CContextOptions(varOptions), 
								pvarResult);
	}

    virtual HRESULT GetContext2Impl(	/*[in,optional]*/ VARIANT varContext,
										/*[in]*/ const CContextOptions& options,
										/*[out,retval]*/ VARIANT* pvarResult)
	{
		if (!m_spiXmlContext) return E_UNEXPECTED;

		return GetContextFromXml(	m_spiXmlContext, 
									varContext, 
									options, 
									pvarResult);
	}

	//
	// SetContext2() is called to set one or more context properties
	//
    STDMETHODIMP SetContext2(	/*[in]*/ VARIANT varContext,
								/*[in]*/ VARIANT varOptions,
								/*[in,optional]*/ VARIANT varValue)
	{
		return SetContext2Impl(	varContext,
								CContextOptions(varOptions), 
								varValue);
	}

    virtual HRESULT SetContext2Impl(	/*[in]*/ VARIANT varContext,
										/*[in]*/ const CContextOptions& options,
										/*[in,optional]*/ VARIANT varValue)
	{
		if (!m_spiXmlContext) return E_UNEXPECTED;

		T* pThis = static_cast<T*>(this);

		return SetContextToXml(	m_spiXmlContext, 
								pThis,
								varContext,
								options, 
								varValue);
	}

	// *******************************************************************
	// CIlxContextImpl Helper Methods
	// *******************************************************************

	//
	// GetStringContext() retrieves the string value of one context
	// property.
	//
    HRESULT GetStringContext(	/*[in]*/ BSTR bstrPropertyName,
								/*[in]*/ BSTR bstrAttributeName,
								/*[out,retval]*/ BSTR* pbstrResult)
	{
		HRESULT hr = S_OK;

		// Verify arguments
		if (!bstrPropertyName || !pbstrResult) return E_POINTER;

		// Sanity check
		if (!m_spiXmlContext) return E_UNEXPECTED;

		// Step 1:	Specify that we want the output to be a BSTR

		CContextOptions options;
		options.eFormat = CTXT_FORMAT_STRING;

		// Step 2:	Set the attribute name if specified.  Otherwise, the "Value"
		//			attribute will be used

		if (bstrAttributeName)
		{
			options.bstrAttributeName = bstrAttributeName;
		}

		// Step 3:	Retrieve the variant value as a BSTR

		CComVariant varValue;
		hr = GetContext2Impl(	CComVariant(bstrPropertyName),
								options, 
								&varValue);
		if (FAILED(hr))
		{
			*pbstrResult = ::SysAllocString(L"");
			return ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to retrieve '%s' property."), 
				bstrPropertyName);
		}

		if (hr == S_FALSE || varValue.vt != VT_BSTR)
		{
			// If the property isn't found then return S_FALSE
			*pbstrResult = ::SysAllocString(L"");
			return S_FALSE;
		}
		else
		{
			// Transfer ownership of the value to the caller
			*pbstrResult = varValue.bstrVal;
			varValue.vt = VT_EMPTY;
			return S_OK;
		}
	}

	//
	// Removes all the context properties
	//
	HRESULT Clear()
	{
		HRESULT hr = S_OK;

		hr = RemoveContextProperty(m_spiXmlContext, bstrNAME_ATTR, 0, 0);
		VERIFY_HR(hr, _T("Failed to clear Context XML document."), ILXLOG_ERROR);

		return S_OK;
	}
	
protected: // DATA

	CComPtr<MSXML::IXMLDOMDocument>		m_spiXmlContext;
};

// Context Helper Functions

HRESULT CreateContextDocument(	MSXML::IXMLDOMDocument** ppiXmlDoc,
								bool bInitialize);
	//
	// Creates an XML document.  If bInitialize is true then a comment
	// is added along with the top level <CONTEXT> document element:
	//
	// <!-- ************************************************************** -->
	// <!-- Copyright (C) 2001-2002 ILX Systems.  All Rights Reserved.     -->
	// <!--                                                                -->
	// <!-- Date: Tuesday, April 16, 2002, 5:34:32 PM                      -->
	// <!--                                                                -->
	// <!-- ************************************************************** -->
	// <Context/>
	//

HRESULT HasContextProperties(MSXML::IXMLDOMDocument* piXmlDoc);
	//
	// Returns S_OK if the document contains one or more PROP
	// elements.  Otherwise returns S_FALSE;
	//

HRESULT CreateXslQueryString(	BSTR bstrAttributeName,
								BSTR bstrAttributeValue,
								TCHAR* szXslQueryString, 
								long nMaxLen);
	//
	// Useful for creating an XSL Query String to be passed to
	// IXMLDOMNode::selectNodes() or IXMLDOMNode::selectSingleNode().
	//
	// If bstrAttributeValue is specified the string looks like:
	//
	//		"Context/Prop[@bstrAttributeName = 'bstrAttributeValue']"
	// 
	// otherwise
	//
	//		"Context/Prop[@bstrAttributeName]"
	// 

HRESULT GetContextFromXml(	/*[in]*/ MSXML::IXMLDOMDocument* piXmlDoc,
							/*[in,optional]*/ VARIANT varContext,
							/*[in]*/ const CContextOptions& options,
							/*[out,retval]*/ VARIANT* pvarResult);
	//
	// Retrieves one or more context values.
	//

HRESULT SetContextToXml(	/*[in,out]*/ MSXML::IXMLDOMDocument* piXmlDoc,
							/*[in]*/ CContextCallback* pContextCallback,
							/*[in]*/ VARIANT varContext,
							/*[in]*/ const CContextOptions& options,
							/*[in,optional]*/ VARIANT varValue);
	//
	// Sets one or more context values.
	//

HRESULT GetContextAttribute(	/*[in]*/ MSXML::IXMLDOMDocument* piXmlDoc,
								/*[in]*/ BSTR bstrPropertyName,
								/*[in]*/ BSTR bstrAttributeName,
								/*[out]*/ VARIANT* pvarAttributeValue);
	//
	// Retrieves an attribute value for a property.
	//

HRESULT RemoveContextProperty(	/*[in,out]*/ MSXML::IXMLDOMDocument* piXmlDoc,
								/*[in]*/ BSTR bstrAttributeName,
								/*[in]*/ BSTR bstrAttributeValue,
								/*[in]*/ BSTR bstrAttributeToRemove);	
	//
	// Removes one or more context properties or attributes.  The attribute
	// name and value (if specified) will be used to find a list of matching
	// properties.  If bstrAttributeToRemove is NULL then all the matching
	// properties will be removed.  Otherwise, only the named attribute will
	// be removed.
	//
	// PARAMS:
	//
	//	piXmlDoc - Context document you want to modify.
	//	bstrAttributeName - Name of the attribute to search for (e.g. "Name",
	//		"WebProp", "Value").
	//	bstrAttributeValue - [Optional] Only remove properties whose attribute
	//		bstrAttributeName has value bstrAttributeValue.  If not specified
	//		then any property with attribute bstrAttributeName will be removed.
	//	bstrAttributeToRemove - [Optional] The attribute to remove.  If NULL,
	//		the property will be removed instead.
	//

bool IsVariantEmpty(const VARIANT* pvar);
	//
	// True if the variant is VT_EMPTY
	//

bool IsVariantUnspecified(const VARIANT* pvar);
	//
	// True if the variant is VT_ERROR or VT_NULL
	//

bool IsVariantEmptyOrUnspecified(const VARIANT* pvar);
	//
	// True if the variant is VT_EMPTY, VT_ERROR, or VT_NULL
	//

#endif // INCLUDED_ILX_CONTEXT_IMPL_H
