///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/AssemblyDiff.cpp 10    5/28/05 1:49p Eaddy $
//
//    File name:	AssemblyDiff.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/AssemblyDiff.cpp $
 * 
 * 10    5/28/05 1:49p Eaddy
 * o Changed copyright
 *
 * Revision 1.4  2005/01/14 23:24:07  me133
 * o Improved logging
 *
 * Revision 1.3  2005/01/14 22:59:39  me133
 * o Trying to debug func ptr problems
 *
 * Revision 1.2  2004/10/26 04:33:04  me133
 * o Fixed bug where methods with exceptions weren't being decoded
 * properly
 *
 * Revision 1.1  2004/10/25 01:03:48  me133
 * o Method IL comparison is now complete and correct
 */

#include "stdafx.h"

#include "AssemblyDiff.h"		// Me

#include "MetadataIterators.h"	// typedef_iterator, etc.

#include "MethodInfo.h"			// CMethodInfo

#include "ILHelper.h"			// opcode_iterator

#include <map>					// STL map<>

using namespace std;

// Type definitions

typedef std::pair<mdToken, mdToken>	TypePairTokens;
typedef std::list<TypePairTokens>	TypeListTokenPairs;

// HELPER FUNCTIONS

 HRESULT StringMetadataDiff(	IMetaDataImport* piImportOrig, 
									IMetaDataImport* piImportNew,
									IMetaDataEmit* piMetaDataEmit, // For emitting the new strings
									TypeListTokenPairs* plistNewOrReorderedStrings);

 HRESULT CreateTypePatches(	typeDef_iterator& itTypeDefOrig, 
									typeDef_iterator& itTypeDefNew,
									const WCHAR* szTypeName,
									ModuleID moduleIDOrig,
									ICorProfilerInfo* piProfInfo,
									IMetaDataEmit* piEmit,
									IMethodMalloc* piAlloc,
									const TypeListTokenPairs& listNewOrReorderedStrings,
									/*[in]*/ bool bFirstTime,
									TypeMapMethodInfo* pmapChangedFunctions);

 HRESULT CreateMethodPatches(	methodDef_iterator& itMethodDefOrig,
									methodDef_iterator& itMethodDefNew,
									const WCHAR* szTypeName,
									ModuleID moduleIDOrig,
									ICorProfilerInfo* piProfInfo,
									IMetaDataEmit* piEmit,
									IMethodMalloc* piAlloc,
									const TypeListTokenPairs& listNewOrReorderedStrings,
									TypeMapMethodInfo* pmapChangedFunctions);

 HRESULT PatchUserStrings(	const TypeMapMethodInfo& listChangedMethods,
									const TypeListTokenPairs& listNewOrReorderedStrings);

 HRESULT PatchLocalVars(		IMetaDataImport* piImportOrig, 
									IMetaDataEmit* piEmit,
									const CMethodInfo* pPatch,
									const CMethodInfo& methodInfoNew);

 bool IsEqualSignature(		PCCOR_SIGNATURE lhs,
									PCCOR_SIGNATURE rhs);
///////////////////
// IMPLEMENTATION
///////////////////

HRESULT CreateUpdatePatches(	/*[in]*/ ICorProfilerInfo* piProfInfo,
								/*[in]*/ ModuleID moduleIDOrig,
								/*[in]*/ IMetaDataImport* piImportNew,
								/*[in]*/ LPCBYTE pbLoadAddressNewModule,
								/*[in]*/ bool bFirstTime,
								/*[out]*/ TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	USES_CONVERSION;

	if (!piProfInfo || !piImportNew || !pbLoadAddressNewModule || !pmapChangedFunctions) 
		return E_POINTER;

	LPCBYTE pbLoadAddressOrigModule = 0;
	WCHAR szModuleName[MAX_PATH] = { 0 };
	ULONG nLen = MAX_PATH;
	hr = piProfInfo->GetModuleInfo(moduleIDOrig, &pbLoadAddressOrigModule, 
										nLen, &nLen, szModuleName, 0);
	VERIFY_HR_COND(hr, pbLoadAddressOrigModule, _T("GetModuleInfo() failed."),
		ILXLOG_ERROR);

	CComPtr<IMetaDataImport> spiImportOrigModule;
	hr = piProfInfo->GetModuleMetaData(	moduleIDOrig, 
										ofRead | ofWrite,
										IID_IMetaDataImport,
										(IUnknown**) &spiImportOrigModule.p);
	VERIFY_HR_COND(hr, spiImportOrigModule.p, 
		_T("GetModuleMetaData() failed."), ILXLOG_ERROR);

	CComPtr<IMetaDataEmit> spiMetaDataEmit;
	hr = spiImportOrigModule->QueryInterface(	IID_IMetaDataEmit, 
												(void**) &spiMetaDataEmit.p );
	VERIFY_HR_COND(hr, spiMetaDataEmit, _T("GetModuleMetaData() failed."), ILXLOG_ERROR);

	CComPtr<IMethodMalloc> spiMethodAlloc;
	hr = piProfInfo->GetILFunctionBodyAllocator(	moduleIDOrig, 
													&spiMethodAlloc.p );
	VERIFY_HR_COND(hr, spiMethodAlloc, _T("GetILFunctionBodyAllocator() failed."),
		ILXLOG_ERROR);

	// Step 1:

	TypeListTokenPairs listNewOrReorderedStrings;
	hr = StringMetadataDiff(	spiImportOrigModule, 
								piImportNew,
								spiMetaDataEmit,
								&listNewOrReorderedStrings);
	VERIFY_HR(hr, _T("Failed to compare string metadata."), ILXLOG_ERROR);

	// Step 1:	Compare CLR build version

	// Step 2:	Compare assembly permissions

	// Step 3:	Compare assembly custom attributes

	// Step 4:	Compare global functions

	// Step 5:	Compare global fields

	// Step 6:	Compare global member refs

	// Step 7:	Create method patches
	//
	//			Iterates over all the types and all the methods of those
	//			types.  If a method is different, a patch is created and
	//			added to pmapChangedFunctions.

	typeDef_iterator itTypeDefOrig(spiImportOrigModule, pbLoadAddressOrigModule);
	typeDef_iterator itTypeDefNew(piImportNew, pbLoadAddressNewModule);

	for(; !itTypeDefOrig.end() && !itTypeDefNew.end(); ++itTypeDefOrig, ++itTypeDefNew)
	{
		_ASSERTE(itTypeDefOrig.getName() != 0);
		if (!itTypeDefOrig.isInterface() && !itTypeDefOrig.isDelegate())
		{
			hr = CreateTypePatches(	itTypeDefOrig, 
									itTypeDefNew, 
									itTypeDefOrig.getName(),
									moduleIDOrig,
									piProfInfo,
									spiMetaDataEmit,
									spiMethodAlloc, 
									listNewOrReorderedStrings,
									bFirstTime,
									pmapChangedFunctions);
			if (FAILED(hr)) return hr;
		}
	}

	// Step 8:	Compare type refs

	// Step 9:	Compare type specs

	// Step 10:	Compare module refs

	// Step 11:	Compare signatures

	// Step 12: Compare extra assembly info (assembly refs, files, exported
	//			types, manifest resources, ?)

	// Step 13: Patch user strings

	try
	{
		hr = PatchUserStrings(	*pmapChangedFunctions,
			listNewOrReorderedStrings);
		if (FAILED(hr)) return hr;
	}
	catch(...)
	{
		ILXLOGMSG0(ILXLOG_ERROR, _T("PatchUserStrings() threw an exception."));
		return E_FAIL;
	}

	return S_OK;
}

HRESULT CreateTypePatches(	typeDef_iterator& itTypeDefOrig, 
							typeDef_iterator& itTypeDefNew,
							const WCHAR* szTypeName,
							ModuleID moduleIDOrig,
							ICorProfilerInfo* piProfInfo,
							IMetaDataEmit* piEmit,
							IMethodMalloc* piAlloc,
							const TypeListTokenPairs& listNewOrReorderedStrings,
							/*[in]*/ bool bFirstTime,
							TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	if (!piAlloc || !pmapChangedFunctions || !piProfInfo) return E_POINTER;

	// Step 1:	Compare type def props

	// Step 2:	Compare class layout

	// Step 3:	Compare fields

	// Step 4:	Compare methods

	methodDef_iterator itMethodDefOrig(itTypeDefOrig);

	if (bFirstTime)
	{
		for(; !itMethodDefOrig.end(); ++itMethodDefOrig)
		{
			CMethodInfo methodInfoOrig;
			hr = methodInfoOrig.init(itMethodDefOrig, szTypeName, false);
			if (FAILED(hr))
				continue;

			// Sanity check: MethodDef tokens are unique
			_ASSERTE(pmapChangedFunctions->find(methodInfoOrig.m_methodDef) ==
				pmapChangedFunctions->end());

			pmapChangedFunctions->insert(make_pair(methodInfoOrig.m_methodDef, methodInfoOrig));
		}

		itMethodDefOrig.reset();
	}

	methodDef_iterator itMethodDefNew(itTypeDefNew);

	for(; !itMethodDefOrig.end() && !itMethodDefNew.end(); ++itMethodDefOrig, ++itMethodDefNew)
	{
		hr = CreateMethodPatches(	itMethodDefOrig, 
									itMethodDefNew, 
									szTypeName,
									moduleIDOrig,
									piProfInfo,
									piEmit,
									piAlloc, 
									listNewOrReorderedStrings,
									pmapChangedFunctions);
		if (FAILED(hr)) 
			continue;
	}

	// Step 5:	Compare properties

	// Step 6:	Compare events

	// Step 7:	Compare method implementations

	// Step 8:	Compare permissions

	// Step 9:	Compare interface implementations

	// Step 10:	Compare custom attributes

	return S_OK;
}

HRESULT CreateMethodPatches(	methodDef_iterator& itMethodDefOrig,
								methodDef_iterator& itMethodDefNew,
								const WCHAR* szTypeName,
								ModuleID moduleIDOrig,
								ICorProfilerInfo* piProfInfo,
								IMetaDataEmit* piEmit,
								IMethodMalloc* piAlloc,
								const TypeListTokenPairs& listNewOrReorderedStrings,
								TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	if (!piAlloc || !pmapChangedFunctions || !piProfInfo) return E_POINTER;

	// Step 1:	Get method info for original method
	//
	//			If the method was patched previously, we should already have
	//			the method info in our list of changed functions.

	CMethodInfo methodInfoOrig;
	CMethodInfo* pMethodInfoOrig = 0;
	bool bAlreadyInList = false;

	// See if we've already patched this method
	TypeMapMethodInfo::iterator it_find = pmapChangedFunctions->find(itMethodDefOrig.getToken());
	if (it_find != pmapChangedFunctions->end())
	{
		pMethodInfoOrig = &(*it_find).second;
		bAlreadyInList = true;
	}

	if (!pMethodInfoOrig)
	{
		// We haven't patched this method so lookup the metadata now
		hr = methodInfoOrig.init(itMethodDefOrig, szTypeName, false);
		if (FAILED(hr)) return hr;

		pMethodInfoOrig = &methodInfoOrig;
	}

	_ASSERTE(pMethodInfoOrig);

	// A PInvoke or Abstract method has no managed bytecode method body
	// and therefore can't be modified by EnC
	if (pMethodInfoOrig->isPInvoke() || pMethodInfoOrig->isAbstract())
		return S_OK;

	// Step 2:	Get method info for the new method

	CMethodInfo methodInfoNew;
	hr = methodInfoNew.init(itMethodDefNew, szTypeName, true);
	if (FAILED(hr)) return hr;

	//ILXLOGMSG2(ILXLOG_SUCCESS, _T("GetMethodPtrInfo('%ls.%ls') succeeded."), 
	//	szTypeName, methodInfoOrig.m_szMemberName);

	// Step 3:	Compare the original and new methods to see if the new method has
	//			been modified
	hr = MethodsAreEquivalent(*pMethodInfoOrig, methodInfoNew, listNewOrReorderedStrings);
	if (FAILED(hr)) return hr;
	if (hr == S_OK) return hr;

	// Methods are different!
	_ASSERTE(hr == S_FALSE);

	//ILXLOGMSG2(ILXLOG_INFO, _T("'%ls.%ls' was modified."), szTypeName, methodInfoNew.m_szMemberName);
	
	// Step 6:	Get the FunctionId for the original method
	//
	//			If JITCompilationFinished event hasn't been received for this method,
	//			i.e. the method hasn't been called yet, the FunctionID will be zero
	//			since the EE only creates them the first time a method is called.

	FunctionID functionID = 0;
	hr = piProfInfo->GetFunctionFromToken(	moduleIDOrig, 
											pMethodInfoOrig->m_methodDef,
											&functionID);
	VERIFY_HR4(hr, ILXLOG_ERROR,
		_T("'%ls.%ls' - GetFunctionFromToken(ModuleID:0x%08X,Def:0x%08X) failed."), 
		szTypeName,
		pMethodInfoOrig->m_szMemberName,
		moduleIDOrig, 
		pMethodInfoOrig->m_methodDef);

	// functionID may be 0 if the function hasn't been called (JIT'd) yet. We log this message later.

	// Step 7:	Create a patch to package the new method

	hr = pMethodInfoOrig->updateMethod(methodInfoNew, moduleIDOrig, functionID, piAlloc);
	if (FAILED(hr)) return hr;

	// Step 10:	Compare params
/*
	paramDef_iterator itParamsLhs(itMethodDefOrig);
	paramDef_iterator itParamsRhs(itMethodDefNew);

	for(; !itParamsLhs.end() && !itParamsRhs.end(); ++itParamsLhs, ++itParamsRhs)
	{
		CParamInfo paramInfoLhs(itParamsLhs);
		CParamInfo paramInfoRhs(itParamsRhs);

		if (paramInfoLhs != paramInfoRhs)
			return S_FALSE;
	}

	if (!itParamsLhs.end() || !itParamsRhs.end())
		return S_FALSE; // Different number of methods
*/
	// Step 11:	Compare custom attributes

	// Step 12:	Compare permissions

	// Step 13:	Compare member refs

	// Step 14:	Compare pinvoke info

	// Step 15:	Patch local variables

	hr = PatchLocalVars(	itMethodDefOrig.getImport(),
							piEmit,
							pMethodInfoOrig,
							methodInfoNew);
	if (FAILED(hr)) return hr;

	// Step 16:	Patch exception handlers (is this needed?)

	// Step 17:	Add to list of functions to patch
	if (!bAlreadyInList)
		pmapChangedFunctions->insert(make_pair(pMethodInfoOrig->m_methodDef, *pMethodInfoOrig));

	return S_OK;
}

HRESULT StringMetadataDiff(	IMetaDataImport* piImportOrig, 
							IMetaDataImport* piImportNew,
							IMetaDataEmit* piMetaDataEmit, // For emitting the new strings
							TypeListTokenPairs* plistNewOrReorderedStrings)
{
	HRESULT hr = S_OK;

	if (!piImportOrig || !piImportNew || !piMetaDataEmit || !plistNewOrReorderedStrings) 
		return E_POINTER;

	typedef map<SStrWLen, mdToken> TypeMapStr2Tok;

	// Map of all the user strings in the original metadata
	TypeMapStr2Tok mapUserString2Tok;

	// Step 1:	Create a map of the original user strings
	//
	//			This has to be done each time since each time
	//			we are emitting new strings.

	userString_iterator itStringsOrig(piImportOrig);

	for(; 
		!itStringsOrig.end(); 
		++itStringsOrig)
	{
		const SStrWLen& str = itStringsOrig.getStr();

		//ILXLOGMSG4(ILXLOG_INFO, _T("STRING: (0x%08X) %ls (Len: %d) (0x%08X)."),
		//	itStringsOrig.getToken(), str.s.c_str(), str.len, mapUserString2Tok[str]);

		_ASSERTE(mapUserString2Tok[str] == 0);
		mapUserString2Tok[str] = itStringsOrig.getToken();
	}

	// Step 2:	Diff the old and new string tables

	ULONG ulStringCountNew = 0;

	for(userString_iterator itStringsNew(piImportNew); 
		!itStringsNew.end(); 
		++itStringsNew)
	{
		++ulStringCountNew;

		const SStrWLen& strUserStringNew = itStringsNew.getStr();
		mdToken tokNew = itStringsNew.getToken();

		// Step 3:	First see if the NEW string exists somewhere
		//			in the ORIG table

		TypeMapStr2Tok::const_iterator it_find =
			mapUserString2Tok.find(strUserStringNew);

		if (it_find != mapUserString2Tok.end())
		{
			// CASE 1: String is in both tables

			// Now see if they are at the same position (token)
			mdToken tokOrig = (*it_find).second;

			if (tokOrig != tokNew)
			{
				// CASE 1a: String is at different locations
				// Just need to fixup the location address
				ILXLOGMSG3(ILXLOG_INFO, _T("FIX STRING ONLY: %ls (0x%08X -> 0x%08X)."),
					strUserStringNew.s.c_str(), tokOrig, tokNew);

				plistNewOrReorderedStrings->push_back(
					TypePairTokens(tokNew, tokOrig));
			}
		}
		else
		{
			// CASE 2:	String is new

			mdString tokNewString = 0;
			hr = piMetaDataEmit->DefineUserString(	strUserStringNew.s.c_str(),
													strUserStringNew.len,
													&tokNewString);
			VERIFY_HR_COND(hr, tokNewString, _T("DefineUserString() failed."), ILXLOG_ERROR);

			if (tokNewString != tokNew)
			{
				// Its very likely that the emitted position will not
				// be the same as the new position
				plistNewOrReorderedStrings->push_back(
					TypePairTokens(tokNew, tokNewString));

				ILXLOGMSG3(ILXLOG_INFO, _T("EMIT STRING AND FIX: %ls (0x%08X->0x%08X)."), 
					strUserStringNew.s.c_str(), tokNew, tokNewString);
			}
			else
			{
				ILXLOGMSG2(ILXLOG_INFO, _T("EMIT STRING ONLY: %ls (0x%08X)."), 
					strUserStringNew.s.c_str(), tokNew);
			}
		}
	}

	return S_OK;
}

HRESULT PatchUserStrings(	const TypeMapMethodInfo& listChangedMethods,
							const TypeListTokenPairs& listNewOrReorderedStrings)
{
	HRESULT hr = S_OK;

	// Step 4: Fixup changed methods to use the correct string table token

	// Iterate over all the changed functions

	for(TypeMapMethodInfo::const_iterator itChangedMethods = listChangedMethods.begin();
		itChangedMethods != listChangedMethods.end();
		++itChangedMethods)
	{
		const CMethodInfo& patch = (*itChangedMethods).second;
		if (!patch.m_bWasModified)
			continue;
		
		// Step 4a:	Enumerate the ldstr opcodes for this method

		opcode_iterator itLDSTRs(	patch.m_pbMethodHeader + patch.m_ulMethodHeaderSize,
									patch.m_pbMethodHeader + patch.m_ulMethodHeaderSize + patch.m_ulCodeSize,
									CEE_LDSTR);
		
		for(; !itLDSTRs.end(); ++itLDSTRs)
		{
			DWORD* pdwStringAddr = (DWORD*) itLDSTRs.getOpCodeArg();
		
			// Step 4b:	Fix the string reference to point to the newly emitted one

			for(TypeListTokenPairs::const_iterator it = listNewOrReorderedStrings.begin();
				it != listNewOrReorderedStrings.end();
				++it)
			{
				const mdString& tokStringFrom = (*it).first;

				if (tokStringFrom == *pdwStringAddr)
				{
					const mdString& tokStringTo = (*it).second;
					*pdwStringAddr = tokStringTo;
					break;
				}
			} // for each reference to rewire

		} // for each ldstr instruction

	} // for each changed function

	return S_OK;
}

HRESULT PatchLocalVars(		IMetaDataImport* piImportOrig, 
							IMetaDataEmit* piMetaDataEmit,
							const CMethodInfo* pPatch,
							const CMethodInfo& methodInfoNew)
{
	HRESULT hr = S_OK;

	if (!piImportOrig || !piMetaDataEmit || !pPatch) return E_POINTER;

	// Step 1:	If there are no local vars then we're done.  This is
	//			always true for tiny methods.

	if (methodInfoNew.m_tokLocalVarSig == mdSignatureNil)
	{
		return S_OK;
	}

	// Step 2:	Do a quick check to see if the local vars are the same

	if (!pPatch->m_pLocalVarSig || !methodInfoNew.m_pLocalVarSig)
	{
		return E_FAIL;
	}

	if (IsEqualSignature(	pPatch->m_pLocalVarSig,
							methodInfoNew.m_pLocalVarSig) &&
		pPatch->m_tokLocalVarSig == methodInfoNew.m_tokLocalVarSig)
	{
		return S_OK;
	}

	// Step 3:	See if the local vars signature is already in the original
	//			assembly in which case we don't have to emit a new signature

	mdSignature sigNew = mdSignatureNil;

	for(signature_iterator itSignaturesOrig(piImportOrig); 
		!itSignaturesOrig.end(); 
		++itSignaturesOrig)
	{
		PCCOR_SIGNATURE pSigOrig = itSignaturesOrig.getSignature();

		if (IsEqualSignature(pSigOrig, methodInfoNew.m_pLocalVarSig))
		{
			// If the signature AND token are equal, just return
			if (itSignaturesOrig.getToken() == methodInfoNew.m_tokLocalVarSig)
			{
				return S_OK;
			}
			else
			{
				// The signatures are the same but the address is different
				// in the new assembly.  Just save the old assemblies address
				// so that we can fixup the new assembly's reference later.
				sigNew = itSignaturesOrig.getToken();

				ILXLOGMSG3(ILXLOG_INFO, _T("FIX LOCAL VAR ONLY: %p 0x%08X -> 0x%08X."),
					methodInfoNew.m_pLocalVarSig, 
					methodInfoNew.m_tokLocalVarSig, 
					sigNew);
				break;
			}
		}
	}

	if (sigNew == mdSignatureNil)
	{
		// Step 4: This is brand new signature so emit it

		ILXLOGMSG1(ILXLOG_INFO, _T("EMIT LOCAL VAR AND FIX: %p."), methodInfoNew.m_pLocalVarSig);

		hr = piMetaDataEmit->GetTokenFromSig(	methodInfoNew.m_pLocalVarSig,
												methodInfoNew.m_ulLocalVarSigSize,
												&sigNew);
		VERIFY_HR_COND(hr, sigNew != 0, _T("GetTokenFromSig() failed."), ILXLOG_ERROR);
	}

	// Step 5:	Update the signature token

	_ASSERTE(sigNew != mdSignatureNil);

	COR_ILMETHOD_FAT* pMethodFat = (COR_ILMETHOD_FAT*) pPatch->m_pbMethodHeader;
	_ASSERTE(pMethodFat);
	_ASSERTE(pMethodFat->IsFat()); // Must be fat since it has local vars
	
	pMethodFat->LocalVarSigTok = sigNew;

	return S_OK;
}

bool IsEqualSignature(		PCCOR_SIGNATURE lhs,
							PCCOR_SIGNATURE rhs)
{
	_ASSERTE(lhs && rhs);
	if (!lhs && !rhs) 
		return true;
	else if (!lhs || !rhs) 
		return false;

	const BYTE typeLhs = *lhs++;
	const BYTE typeRhs = *rhs++;

	if (typeLhs != typeRhs)
		return false;

	BYTE countLhs = *lhs++;
	BYTE countRhs = *rhs++;

	if (countLhs != countRhs)
		return false;

	for(int i = 0; i < countLhs; ++i)
	{
		if (lhs[i] != rhs[i])
			return false;
	}

	return true;
}
