///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MetadataHelpers.cpp 2     5/28/05 1:20p Eaddy $
//
//    File name:	MetadataHelpers.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MetadataHelpers.cpp $
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.7  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.6  2004/10/24 19:33:10  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.5  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.4  2004/10/11 23:22:52  me133
 * o Patch now contains method name
 * o We now skip over methods we couldn't lookup (instead of crashing)
 * o Uses IsBadReadPtr() to avoid access violations when we fail to lookup
 * a method
 *
 * Revision 1.3  2004/10/10 07:40:51  me133
 * o Converted print statements to logging
 *
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "stdafx.h"

#include "MetadataHelpers.h"	// Me

#include <cor.h>				// IMAGE_DIRECTORY_ENTRY_COMHEADER

#include <dbghelp.h>			// ImageNtHeader()

// MACROS

#define VAL32(x) x // Only works for little endian architectures

///////////////////
// IMPLEMENTATION
///////////////////

// This function returns the address to the MapView of file and file size.
// Adapted from MetaInfo .NET sample.
HRESULT MyGetMapViewOfFile(	const char* szFile, 
							PBYTE *ppbMap, 
							DWORD *pdwFileSize)
{
	HRESULT hr = S_OK;

	if (!szFile || !*szFile) return E_POINTER;
	
	// Must specify one or the other
	if (!ppbMap && !pdwFileSize) return E_POINTER;

    HANDLE hFile = CreateFileA(szFile,
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL); 
	VERIFY_COND1(hFile != INVALID_HANDLE_VALUE, ILXLOG_ERROR, _T("CreateFileA() failed: '%s'."), 
		szFile);

	if (pdwFileSize)
	{
	    DWORD dwHighSize = 0;
		*pdwFileSize = ::GetFileSize(hFile, &dwHighSize);
		DWORD dwLastError = ::GetLastError();
		if (*pdwFileSize == 0xFFFFFFFF && dwLastError != NO_ERROR)
		{
			::CloseHandle(hFile);
			return ILXLOG1(HRESULT_FROM_WIN32(dwLastError), ILXLOG_ERROR,
				_T("GetFileSize() failed: '%s'."), szFile);
		}

		_ASSERTE(dwHighSize == 0); // We don't support very big files
	}

	if (!ppbMap)
	{
		::CloseHandle(hFile);
		return S_OK;
	}

	HANDLE hMapFile = ::CreateFileMappingA(	hFile, 
											NULL, 
											PAGE_READONLY, 
											0, 
											0, 
											NULL);
	::CloseHandle(hFile);
	hFile = 0; // So we don't accidentally use it

	VERIFY_COND1(hMapFile, ILXLOG_ERROR, _T("CreateFileMappingA() failed: '%s'."), szFile);

    *ppbMap = (PBYTE) MapViewOfFile(hMapFile, FILE_MAP_READ, 0, 0, 0);
	::CloseHandle(hMapFile);
	hMapFile = 0;

	VERIFY_COND1(*ppbMap, ILXLOG_ERROR, _T("MapViewOfFile() failed: '%s'."), szFile);

	return S_OK;

} // void GetMapViewOfFile()

// Adapted from:
//		Roshan James, rosh_spark@gmx.net
//		May 2003
//		http://www27.brinkster.com/sparksite/work/trashbin/trashbin.htm
//
// ME: I had to modify his code to call ImageRvaToVa().  For some
// reason I couldn't get the FindObjMetaData in mdobj.cpp
// (in MetaInfo sample) to work (access violations).  I think it
// only works for .LIB files that have OBJ sections or something.
// I need it to work for EXEs and DLLs that have metadata.
//
HRESULT MyFindMetaData(	PVOID pImage, 
						DWORD dwFileLength,
						PVOID* ppMetaData, 
						ULONG* pcbMetaData)
{
	// Verify arguments
	if (!pImage || !ppMetaData || !pcbMetaData) return E_POINTER;

    // Get a pointer to the header and the first section.

    // Initialize out params
    *ppMetaData = NULL;
    *pcbMetaData = 0;

	_IMAGE_DOS_HEADER* pDosH = (_IMAGE_DOS_HEADER*) pImage;
	char* pPESign = (char*)pImage + pDosH->e_lfanew;	//e_lfanew contains the address of the rest of the header

	if(!(	pPESign[0] == 'P' &&  
			pPESign[1] == 'E' &&  
			pPESign[2] == 0 &&  
			pPESign[3] == 0	))
	{
		ILXLOGMSG(_T("This is not a PE file according to the file signature."),
			ILXLOG_ERROR);;
		return E_FAIL;
	}

	_IMAGE_FILE_HEADER* pCoffH = (_IMAGE_FILE_HEADER*)(pPESign + 4) ; 	//4 byte PE signature
	_IMAGE_OPTIONAL_HEADER* pOptH = (_IMAGE_OPTIONAL_HEADER*)((char *)pCoffH + sizeof(_IMAGE_FILE_HEADER)); //these lie one after the other

	PIMAGE_NT_HEADERS pNtHeader = ::ImageNtHeader(pImage);
	if (!pNtHeader)
	{
		HRESULT hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("ImageNtHeader failed."), ILXLOG_ERROR);
	}

	//these are content specific to a managed file
	_IMAGE_DATA_DIRECTORY* pDDCorHdr = &(pOptH->DataDirectory[IMAGE_DIRECTORY_ENTRY_COMHEADER]);
	if (!pDDCorHdr) return E_FAIL;

	//if this file is a managed file then this entry should exist
	IMAGE_COR20_HEADER* pCorHdr = 
		(IMAGE_COR20_HEADER*) ImageRvaToVa(	pNtHeader, 
											pImage, 
											VAL32(pDDCorHdr->VirtualAddress), 
											0);
	if (!pCorHdr) return E_FAIL;

	// KLUDGE: Sometimes the CLR treates the RVA as a virtual address (VA) and
	// sometimes as an offset.  For example, if the CLR is loading
	// assemblies it seems to treat rva as an offset. If we are mapping
	// PE files directly, the rva appears to be a VA.  If ImageRvaToVa()
	// doesn't appear to work we try treating it as an offset.  This is
	// very kludgy but I don't know another way to do this.
	// See CorMap::ReadHeaders() in cormap.cpp in Rotor to see an example.
	if (pCorHdr->MajorRuntimeVersion == 0)
	{
		IMAGE_COR20_HEADER* pCorHdr2 = 
			(IMAGE_COR20_HEADER*) pImage + VAL32(pDDCorHdr->VirtualAddress);
		if (pCorHdr->MajorRuntimeVersion != 0)
		{
			pCorHdr = pCorHdr2;
		}
	}

	*ppMetaData = (PVOID) ImageRvaToVa(		pNtHeader, 
											pImage, 
											VAL32(pCorHdr->MetaData.VirtualAddress), 
											0);
	*pcbMetaData = pCorHdr->MetaData.Size;

	return S_OK;
}
