///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MetadataIterators.cpp 6     5/28/05 1:20p Eaddy $
//
//    File name:	MetadataIterators.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MetadataIterators.cpp $
 * 
 * 6     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "StdAfx.h"

#include "MetadataIterators.h"	// Me

#include <corerror.h>			// CLDB_S_TRUNCATION

/////////////////////////////////////
// TYPE DEF ITERATOR IMPLEMENTATION
/////////////////////////////////////

typeDef_iterator::typeDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress)
	: _Mybase(piImport, pbLoadAddress)
	, m_szTypeName()
	, m_dwFlags(0)
	, m_tokBaseType(mdTokenNil)
{ 
	getNextBatch();
}

void typeDef_iterator::initCurrent()
{
	getProps(this->operator*(), m_szTypeName, (sizeof(m_szTypeName)/sizeof(*m_szTypeName))-1, &m_dwFlags, &m_tokBaseType);
}

HRESULT typeDef_iterator::getProps(mdToken tokType, WCHAR* szTypeName, ULONG nMaxLen, DWORD* pdwFlags, mdToken* ptokBaseType) const
{
	HRESULT hr = S_OK;

	szTypeName[0] = L'\0';

	ULONG cchTypeName = 0;
	if (TypeFromToken(tokType) == mdtTypeRef)
	{
		CComPtr<IMetaDataImport> spiImportType;

		hr = m_spiImport->ResolveTypeRef(tokType, IID_IMetaDataImport, (IUnknown**)(&spiImportType.p), &tokType); 
		VERIFY_HR1(hr, ILXLOG_ERROR, _T("ResolveTypeRef(0x%08X) failed."), tokType);
		
		hr = spiImportType->GetTypeDefProps(	tokType, 
												szTypeName, 
												nMaxLen, 
												&cchTypeName, 
												pdwFlags,
												ptokBaseType);
	}
	else 
	{
		// This is expensive
		hr = m_spiImport->GetTypeDefProps(	tokType, 
											szTypeName, 
											nMaxLen, 
											&cchTypeName, 
											pdwFlags, 
											ptokBaseType);
	}

	VERIFY_HR1(hr, ILXLOG_ERROR, _T("GetTypeDefProps(0x%08X) failed."), tokType);
	if (cchTypeName > nMaxLen)
	{
		return ILXLOG3(E_UNEXPECTED, ILXLOG_ERROR, _T("Type name for 0x%08X too big (%d), max is %d."), tokType, cchTypeName, nMaxLen);
	}

	// For some reason we can't retrieve the type name for System.Object
	// JScript.NET doesn't have types
	//_ASSERTE(*szTypeName || (cchTypeName == 0 && hr == S_FALSE));

	return hr;
}

bool typeDef_iterator::isDelegate() const
{
	HRESULT hr = S_OK;

	// Traverse up the inheritence hierarchy to see if we are derived from
	// System.MulticastDelegate

	mdToken tokBaseType = m_tokBaseType;

	while (	hr == S_OK && 
			TypeFromToken(tokBaseType) != mdTokenNil && 
			!IsNilToken(tokBaseType))
	{
		const int nMaxLen = 2048; // 2048 from alink.h
		WCHAR szBaseTypeName[nMaxLen+1]; 

		hr = getProps(tokBaseType, szBaseTypeName, nMaxLen, 0, &tokBaseType);

		// An optimization would be to store this result. However, isDelegate() is only
		// ever called once for the current iteration
		if (hr == S_OK && wcscmp(szBaseTypeName, L"System.MulticastDelegate") == 0)
			return true;
	}
	
	return false;
}

bool typeDef_iterator::isInterface() const
{
	return IsTdInterface(m_dwFlags);
}

const WCHAR* typeDef_iterator::getName() const
{
	return m_szTypeName;
}

void typeDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumTypeDefs( &m_hEnum,
									m_ary, 
									NumItems(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumTypeDefs() failed."), ILXLOG_ERROR);
}

///////////////////////////////////////
// METHOD DEF ITERATOR IMPLEMENTATION
///////////////////////////////////////

methodDef_iterator::methodDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress, mdTypeDef typeDef) 
	: _Mybase(piImport, pbLoadAddress)
	, m_typeDef(typeDef)
{
	getNextBatch();
}

methodDef_iterator::methodDef_iterator(typeDef_iterator& itTypeDef) 
	: _Mybase(itTypeDef)
	, m_typeDef(*itTypeDef)
{
	getNextBatch();
}

void methodDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumMethods( &m_hEnum,
									m_typeDef,
									m_ary, 
									NumItems(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumMethods() failed."), ILXLOG_ERROR);
}

//////////////////////////////////////
// PARAM DEF ITERATOR IMPLEMENTATION
//////////////////////////////////////

paramDef_iterator::paramDef_iterator(IMetaDataImport* piImport, LPCBYTE pbLoadAddress, mdMethodDef methodDef) 
	: _Mybase(piImport, pbLoadAddress)
	, m_methodDef(methodDef)
{
	getNextBatch();
}

paramDef_iterator::paramDef_iterator(methodDef_iterator& itMethodDef) 
	: _Mybase(itMethodDef)
	, m_methodDef(*itMethodDef)
{
	getNextBatch();
}

void paramDef_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumParams(	&m_hEnum,
									m_methodDef,
									m_ary, 
									NumItems(m_ary), 
									&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumParams() failed."), ILXLOG_ERROR);
}

////////////////////////////////////////
// USER STRING ITERATOR IMPLEMENTATION
////////////////////////////////////////

userString_iterator::userString_iterator(IMetaDataImport* piImport) 
	: _Mybase(piImport)
	, m_pszTemp(0)
	, m_ulSize(0)
{
	reserve(nINITIAL_SIZE);
	getNextBatch();
}

userString_iterator::~userString_iterator()
{
	free(m_pszTemp);
}

// The string returned will be invalid if you increment or
// destroy the iterator
const SStrWLen& userString_iterator::getStr() const
{
	static SStrWLen strTemp;

	HRESULT hr = S_OK;

	if (end())
	{
		_ASSERTE(!"End of iteration");
		strTemp.s.clear();
		strTemp.len = -1;
		return strTemp;
	}

	ULONG chUserString = 0;           // Size of user string.

    while (true) 
	{ 
		// Try to get the string into the existing buffer.
		// Subtract 1 because we need the extra space for the
		// null terminator
        hr = m_spiImport->GetUserString(	this->operator*(), 
											m_pszTemp,
											m_ulSize-1, 
											&chUserString);
        if (hr == CLDB_S_TRUNCATION)
        {   
			// Buffer wasn't big enough, try to enlarge it.
			reserve(chUserString);
        }
		else
		{
			// Buffer was big enough
			break;
		}
    }

	if (m_pszTemp && chUserString >= 0 && chUserString < m_ulSize)
	{
		m_pszTemp[chUserString] = L'\0';
	}

	strTemp.s = m_pszTemp;
	strTemp.len = chUserString;

	CHECK_HR(hr, _T("GetUserString() failed."), ILXLOG_ERROR);

	return strTemp;
}

void userString_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumUserStrings(	&m_hEnum,
										m_ary, 
										NumItems(m_ary), 
										&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumUserStrings() failed."), ILXLOG_ERROR);
}

void userString_iterator::reserve(ULONG nCapacity) const
{
	// Add one for the null terminator which metadata doesn't set
	nCapacity = max(nCapacity, m_ulSize * nINC_FACTOR) + 1;

	// Buffer wasn't big enough, try to enlarge it
	free(m_pszTemp);
	m_pszTemp = (WCHAR*) malloc(nCapacity * sizeof(WCHAR));
	m_ulSize = nCapacity;
}

//////////////////////////////////////
// SIGNATURE ITERATOR IMPLEMENTATION
//////////////////////////////////////

signature_iterator::signature_iterator(IMetaDataImport* piImport)
	: _Mybase(piImport)
{ 
	getNextBatch();
}

void signature_iterator::getNextBatchImpl()
{
	HRESULT hr = S_OK;

	hr = m_spiImport->EnumSignatures(	&m_hEnum,
										m_ary, 
										NumItems(m_ary), 
										&m_ulCount);
	VERIFY_HR_COND_V(hr, m_ulCount >= 0, _T("EnumSignatures() failed."), ILXLOG_ERROR);
}

PCCOR_SIGNATURE signature_iterator::getSignature() const
{
	if (end() || !m_spiImport)
	{
		return 0;
	}
	else
	{
		PCCOR_SIGNATURE pSig = 0;
		ULONG ulSigSize = 0;

		HRESULT hr = m_spiImport->GetSigFromToken(	this->operator*(),
													&pSig,
													&ulSigSize);
		VERIFY_HR_(0, hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);

		return pSig;
	}
}
