///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MethodInfo.cpp 4     5/28/05 1:20p Eaddy $
//
//    File name:	MethodInfo.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MethodInfo.cpp $
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2005/01/14 23:28:47  me133
 * o Now initialized method name
 *
 * Revision 1.3  2005/01/14 23:03:47  me133
 * o Made some functions static
 * o Move check into separate function
 *
 * Revision 1.2  2004/10/26 04:43:07  me133
 * no message
 *
 * Revision 1.1  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 */

#include "stdafx.h"

#include "MethodInfo.h"			// Me

#include "ILHelper.h"			// opcode_iterator, FindOpCode()

#include <dbghelp.h>			// ImageNtHeader()

using namespace std;

// HELPER FUNCTIONS

 HRESULT GetMethodIL(	/*[in]*/  unsigned __int8* const pbMethodHeader,
						/*[out]*/ ULONG* pulMethodSize,
						/*[out]*/ ULONG* pulHeaderSize,
						/*[out]*/ LPCBYTE* ppbMethodIL,
						/*[out]*/ ULONG* pulCodeSize,
						/*[out]*/ mdSignature* ptokLocalVarSig);

 HRESULT GetMethodHeader(	LPCBYTE pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							unsigned __int8** ppbMethodHeader);

 HRESULT GetMethodPtrInfo(	LPCBYTE pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							ULONG* pulMethodSize,
							unsigned __int8** ppbMethodHeader,
							ULONG* pulMethodHeaderSize,
							LPCBYTE* ppbMethodIL,
							ULONG* pulCodeSize,
							mdSignature* ptokLocalVarSig);

 HRESULT MethodILIsEquivalent(	const CMethodInfo& pre, 
										const CMethodInfo& post, 
										const TypeListTokenPairs& listEquivalentStringTokens);

 bool IsBadMethodHeaderPtr(unsigned __int8* pbMethodHeader);

///////////////////////////////
// METHOD INFO IMPLEMENTATION
///////////////////////////////

CMethodInfo::CMethodInfo()
	: m_spiImport(0)
	, m_typeDef(mdTypeDefNil)
	, m_methodDef(mdMethodDefNil)
	, m_szTypeName()
	, m_szMemberName()
	, m_ulNameLen(0)
	, m_dwFlags(0)
	, m_pbSigBlob(0)
	, m_ulSigBlob(0)
	, m_ulCodeRVA(0)
	, m_ulImplFlags(0)
	, m_pbMethodHeader(0)
	, m_ulMethodHeaderSize(0)
	, m_pbMethodIL(0)
	, m_ulCodeSize(0)
	, m_tokLocalVarSig(mdSignatureNil)
	, m_pLocalVarSig(0)
	, m_ulLocalVarSigSize(0)
	, m_moduleID(-1)
	, m_functionID(-1)
	, m_bWasModified(false)
{ 
	*m_szTypeName = '\0';
	*m_szMemberName = '\0';
}

CMethodInfo::CMethodInfo(const CMethodInfo& that)
	: m_spiImport(that.m_spiImport)
	, m_typeDef(that.m_typeDef)
	, m_methodDef(that.m_methodDef)
	, m_szTypeName()
	, m_szMemberName()
	, m_ulNameLen(that.m_ulNameLen)
	, m_dwFlags(that.m_dwFlags)
	, m_pbSigBlob(that.m_pbSigBlob)
	, m_ulSigBlob(that.m_ulSigBlob)
	, m_ulCodeRVA(that.m_ulCodeRVA)
	, m_ulImplFlags(that.m_ulImplFlags)
	, m_pbMethodHeader(that.m_pbMethodHeader)
	, m_ulMethodHeaderSize(that.m_ulMethodHeaderSize)
	, m_pbMethodIL(that.m_pbMethodIL)
	, m_ulCodeSize(that.m_ulCodeSize)
	, m_tokLocalVarSig(that.m_tokLocalVarSig)
	, m_pLocalVarSig(that.m_pLocalVarSig)
	, m_ulLocalVarSigSize(that.m_ulLocalVarSigSize)
	, m_moduleID(that.m_moduleID)
	, m_functionID(that.m_functionID)
	, m_bWasModified(that.m_bWasModified)
{
	wcscpy(m_szTypeName, that.m_szTypeName);
	wcscpy(m_szMemberName, that.m_szMemberName);
}

HRESULT CMethodInfo::init(methodDef_iterator& itMethodDef, const WCHAR* szTypeName, bool bIsImage)
{
	return init(itMethodDef.getImport(), *itMethodDef, szTypeName, itMethodDef.getLoadAddress(), bIsImage);
}

HRESULT CMethodInfo::init(ICorProfilerInfo* piProfInfo, FunctionID functionID, const WCHAR* szTypeName, LPCBYTE pbLoadAddress, bool bIsImage)
{
	HRESULT hr = S_OK;

	if (!piProfInfo) return E_POINTER;

	m_functionID = functionID;

    mdToken methodDef = mdMethodDefNil;
    CComPtr<IMetaDataImport> spiMetaDataImport;

    hr = piProfInfo->GetTokenAndMetaDataFromFunction( functionID, 
									IID_IMetaDataImport,
									(LPUNKNOWN *)&spiMetaDataImport.p, 
									&methodDef );
    if (FAILED(hr)) return hr;
	if (!spiMetaDataImport) return E_UNEXPECTED;

	return init(spiMetaDataImport, methodDef, szTypeName, pbLoadAddress, bIsImage);
}

HRESULT CMethodInfo::init(IMetaDataImport* piImport, mdMethodDef methodDef, const WCHAR* szTypeName, LPCBYTE pbLoadAddress, bool bIsImage)
{
	HRESULT hr = S_OK;

	// No AddRef/Release necessary
	_ASSERTE(piImport);
	if (!piImport) return E_POINTER;

	_ASSERTE(!m_spiImport); // Only call init() once

	m_spiImport = piImport;
	m_methodDef = methodDef;

	if (szTypeName)
	{
		_ASSERTE(*szTypeName);
		wcscpy(m_szTypeName, szTypeName);
	}

	// Step 1:	Get some method properties

	hr = piImport->GetMethodProps(	methodDef, 
									&m_typeDef, 
									m_szMemberName, 
									nMAX_ID,
									&m_ulNameLen, 
									&m_dwFlags, 
									&m_pbSigBlob, 
									&m_ulSigBlob, 
									&m_ulCodeRVA, 
									&m_ulImplFlags);
	VERIFY_HR(hr, _T("GetMethodProps() failed."), ILXLOG_ERROR);

	if (!pbLoadAddress) return S_FALSE;

	// PInvoke methods are those defined using dllimport/extern. We ignore
	// them since they aren't managed code and don't have bytecode.
	if (isPInvoke()) return S_FALSE;

	// Abstract methods don't have a method body.
	if (isAbstract()) return S_FALSE;

	// Step 2:	Get the method header and IL code info

	// This is expensive!
	hr = GetMethodPtrInfo(	pbLoadAddress,
							m_ulCodeRVA,
							bIsImage,
							&m_ulMethodSize,
							&m_pbMethodHeader,
							&m_ulMethodHeaderSize,
							&m_pbMethodIL,
							&m_ulCodeSize,
							&m_tokLocalVarSig);
	VERIFY_HR2(hr, ILXLOG_ERROR, _T("GetMethodPtrInfo('%ls.%ls') failed."), 
		getTypeName(), m_szMemberName);

	// Step 3:	Get the local variable signatures

	if (m_tokLocalVarSig != mdSignatureNil)
	{
		hr = piImport->GetSigFromToken(	m_tokLocalVarSig,
										&m_pLocalVarSig,
										&m_ulLocalVarSigSize);
		VERIFY_HR(hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);
	}

	return S_OK;
}

HRESULT CMethodInfo::updateMethod(	const CMethodInfo& methodInfoNew, 
									ModuleID moduleID,
									FunctionID functionID,
									IMethodMalloc* piAlloc)
{
	HRESULT hr = S_OK;

	//m_spiImport = that.m_spiImport;

	_ASSERTE(m_typeDef == methodInfoNew.m_typeDef);
	_ASSERTE(m_methodDef == methodInfoNew.m_methodDef);
	_ASSERTE(wcscmp(m_szTypeName, methodInfoNew.m_szTypeName) == 0);
	_ASSERTE(wcscmp(m_szMemberName, methodInfoNew.m_szMemberName) == 0);
	_ASSERTE(m_ulNameLen == methodInfoNew.m_ulNameLen);

	m_dwFlags = methodInfoNew.m_dwFlags;
	
	//m_pbSigBlob = methodInfoNew.m_pbSigBlob; // Don't want stray pointers
	
	m_ulSigBlob = methodInfoNew.m_ulSigBlob; // Dangerous?
	m_ulCodeRVA = methodInfoNew.m_ulCodeRVA; // Dangerous?
	
	m_ulImplFlags = methodInfoNew.m_ulImplFlags;

	_ASSERTE(methodInfoNew.m_pbMethodHeader);

	// Step 4:	Allocate memory for the new method
	m_pbMethodHeader = (LPBYTE) piAlloc->Alloc(methodInfoNew.m_ulMethodSize);
	VERIFY_COND(m_pbMethodHeader, _T("Alloc() failed."), ILXLOG_ERROR);

	// Step 5:	Copy the method (header and code)
	memcpy((void*)m_pbMethodHeader, methodInfoNew.m_pbMethodHeader, 
		methodInfoNew.m_ulMethodSize);

	hr = GetMethodIL(	/*[in]*/  m_pbMethodHeader,
						/*[out]*/ &m_ulMethodSize,
						/*[out]*/ &m_ulMethodHeaderSize,
						/*[out]*/ &m_pbMethodIL,
						/*[out]*/ &m_ulCodeSize,
						/*[out]*/ &m_tokLocalVarSig);

	_ASSERTE(m_ulMethodSize = methodInfoNew.m_ulMethodSize);
	_ASSERTE(m_ulMethodHeaderSize = methodInfoNew.m_ulMethodHeaderSize);
	_ASSERTE(m_ulCodeSize == methodInfoNew.m_ulCodeSize);

	if (m_tokLocalVarSig != mdSignatureNil)
	{
		hr = m_spiImport->GetSigFromToken(	m_tokLocalVarSig,
											&m_pLocalVarSig,
											&m_ulLocalVarSigSize);
		VERIFY_HR(hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);
	}

	_ASSERTE(m_ulLocalVarSigSize == methodInfoNew.m_ulLocalVarSigSize);

	m_moduleID = moduleID;
	m_functionID = functionID;
	m_bWasModified = true;

	return S_OK;
}

LPBYTE CMethodInfo::getCodeStart() const
{
	_ASSERTE(m_typeDef != mdTypeDefNil);
	return const_cast<LPBYTE>(m_pbMethodIL);
}

LPBYTE CMethodInfo::getCodeEnd() const
{
	_ASSERTE(m_typeDef != mdTypeDefNil);
	return getCodeStart() + m_ulCodeSize;
}

bool CMethodInfo::isPInvoke() const
{
	return IsMdPinvokeImpl(m_dwFlags) != 0;
}

bool CMethodInfo::isAbstract() const
{
	return IsMdAbstract(m_dwFlags) != 0;
}

const WCHAR* CMethodInfo::getTypeName() /* const */
{
	HRESULT hr = S_OK;

	// For Global methods, typedef is nil. I've noticed
	// this with unmanaged/pinvoke methods.
	//_ASSERTE(m_typeDef != mdTypeDefNil);

	if (m_typeDef == mdTypeDefNil)
	{
		return L"[global]";
	}
	else if (!*m_szTypeName && m_spiImport)
	{
		DWORD dwSize = nMAX_ID-1;
		DWORD dwRetSize = dwSize;

		hr = m_spiImport->GetTypeDefProps(	m_typeDef, 
											m_szTypeName, 
											dwSize,
											&dwRetSize, 
											0, 0 );
		VERIFY_HR1_(L"<ERROR>", hr, ILXLOG_WARNING,
			_T("Failed to obtain containing type for method '%ls'."),
			m_szMemberName);

		_ASSERTE(dwRetSize <= dwSize);
	}

	return m_szTypeName;
}

/////////////////////
// HELPER FUNCTIONS
/////////////////////

HRESULT MethodsAreEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenPairs& listEquivalentStringTokens)
{
	HRESULT hr = S_OK;

	// Ignore m_typeDef differences due to different assemblies
	// Ignore m_ulCodeRVA differences due to different assemblies

	// Do quick comparisons first

	if (pre.m_szMemberName != post.m_szMemberName)
	{
		if (!pre.m_szMemberName || !post.m_szMemberName)
		{
			ILXLOGMSG2(ILXLOG_WARNING, _T("Method names not equal: '%ls' != '%ls'."), 
				pre.m_szMemberName ? pre.m_szMemberName : L"(null)", 
				post.m_szMemberName ? post.m_szMemberName : L"(null)");
			return E_FAIL; // One is null but not the other
		}

		if (wcscmp(pre.m_szMemberName, post.m_szMemberName) != 0)
		{
			ILXLOGMSG2(ILXLOG_WARNING, _T("Method names not equal: '%ls' != '%ls'."), 
				pre.m_szMemberName, post.m_szMemberName);
			return E_FAIL; // Strings are different (method name has changed)
		}
	}

	if (pre.m_ulNameLen != post.m_ulNameLen)
	{
		ILXLOGMSG2(ILXLOG_WARNING, _T("Method names not equal: '%ls' != '%ls'."), 
			pre.m_szMemberName ? pre.m_szMemberName : L"(null)", 
			post.m_szMemberName ? post.m_szMemberName : L"(null)");
		return E_FAIL; // Method names aren't the same
	}

	if (pre.m_dwFlags != post.m_dwFlags)
	{
		ILXLOGMSG3(ILXLOG_INFO, _T("Flags have changed for method '%ls' (Old:%d New:%d)."), 
			pre.m_szMemberName, pre.m_dwFlags, post.m_dwFlags);
		return S_FALSE;
	}

	if (pre.m_ulSigBlob != post.m_ulSigBlob)
	{
		ILXLOGMSG3(ILXLOG_INFO, _T("Signature size has changed for method '%ls' (Old:%d New:%d)."), 
			pre.m_szMemberName, pre.m_ulSigBlob, post.m_ulSigBlob);
		return S_FALSE;
	}

	if (pre.m_ulImplFlags != post.m_ulImplFlags)
	{
		ILXLOGMSG3(ILXLOG_INFO, _T("Impl flags have changed for method '%ls' (Old:%d New:%d)."), 
			pre.m_szMemberName, pre.m_ulImplFlags, post.m_ulImplFlags);
		return S_FALSE;
	}

	if (pre.m_ulMethodHeaderSize != post.m_ulMethodHeaderSize)
	{
		ILXLOGMSG3(ILXLOG_INFO, _T("Header size has changed for method '%ls' (Old:%d New:%d)."), 
			pre.m_szMemberName, pre.m_ulMethodHeaderSize, post.m_ulMethodHeaderSize);
		return S_FALSE;
	}

	if (pre.m_ulCodeSize != post.m_ulCodeSize)
	{
		ILXLOGMSG3(ILXLOG_INFO, _T("Code size has changed for method '%ls' (Old:%d New:%d)."), 
			pre.m_szMemberName, pre.m_ulCodeSize, post.m_ulCodeSize);
		return S_FALSE;
	}

	// Compare local variables

	if (!pre.m_pbMethodHeader || !post.m_pbMethodHeader)
	{
		// Unable to obtain method header for some reason
		ILXLOGMSG3(ILXLOG_WARNING, _T("Invalid method header for '%ls' (Old:0x%08X New:%0x%08X)."), 
			pre.m_szMemberName, pre.m_pbMethodHeader, post.m_pbMethodHeader);
		return E_FAIL; 
	}

	COR_ILMETHOD_DECODER methodDecoderPre( (COR_ILMETHOD*) pre.m_pbMethodHeader );
	COR_ILMETHOD_DECODER methodDecoderPost( (COR_ILMETHOD*) post.m_pbMethodHeader );

	mdSignature sigLocalPre = methodDecoderPre.GetLocalVarSigTok() ? 
		methodDecoderPre.GetLocalVarSigTok() : mdSignatureNil;
	mdSignature sigLocalPost = methodDecoderPost.GetLocalVarSigTok() ?
		methodDecoderPost.GetLocalVarSigTok() : mdSignatureNil;

	if (sigLocalPre != sigLocalPost)
	{
		// Actually, we should call GetSigFromToken() to compare the actual
		// contents of the local vars.  However, this is probably overkill
		// since we are doing an IL code comparison below.
		ILXLOGMSG1(ILXLOG_INFO, _T("Local variables have changed for '%ls'."), 
			pre.m_szMemberName);
		return S_FALSE;
	}

	// Compare method header

	if (pre.m_pbMethodHeader != post.m_pbMethodHeader)
	{
		if (!pre.m_pbMethodHeader || !post.m_pbMethodHeader)
		{
			ILXLOGMSG3(ILXLOG_WARNING, _T("Invalid method header for '%ls' (Old:0x%08X New:%0x%08X)."), 
				pre.m_szMemberName, pre.m_pbMethodHeader, post.m_pbMethodHeader);
			return E_FAIL; // One is null but not the other
		}

		if (memcmp(pre.m_pbMethodHeader, post.m_pbMethodHeader, pre.m_ulMethodHeaderSize) != 0)
		{
			ILXLOGMSG1(ILXLOG_WARNING, _T("Method headers are different for '%ls'."), 
				pre.m_szMemberName);
			return E_FAIL; // Method headers should be the same
		}
	}

	// Compare the method IL
	hr = MethodILIsEquivalent(pre, post, listEquivalentStringTokens);
	VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to compare method '%ls'."),
		pre.m_szMemberName);

	if (hr == S_FALSE)
	{
		ILXLOGMSG1(ILXLOG_INFO, _T("Method IL is different for '%ls'."), 
			pre.m_szMemberName);
		return S_FALSE;
	}

	// Compare exception handlers

	return S_OK;
}

// The methods might only be different in terms of token references
// (strings vars) which end up getting fixed up later. The method
// bodies may be exactly the same but the actual strings referenced
// are different. This will happen if all you do is change a string
// value in a method.
//
// We must "smartly" compare IL in the face of 1) string refs that
// seem different but are really the same, 2) string refs that
// seem the same but are really different.  We only have to
// perform this special comparison if this list of mapped references
// is not empty. If its empty, a simple binary compare will do.
HRESULT MethodILIsEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenPairs& listEquivalentStringTokens)
{
	if (pre.m_pbMethodIL == post.m_pbMethodIL)
	{
		return E_UNEXPECTED;
	}
	else if (!pre.m_pbMethodIL || !post.m_pbMethodIL)
	{
		return E_FAIL; // One is null but not the other
	}
	else if (listEquivalentStringTokens.size() == 0)
	{
		// If the string tables are exactly the same then we can just perform
		// a binary comparison
		return memcmp(	pre.m_pbMethodIL, 
						post.m_pbMethodIL, 
						pre.m_ulCodeSize) == 0 ? S_OK : S_FALSE;
	}

	// Compare each method instruction

	opcode_iterator itOPCODEs_PRE(pre);
	opcode_iterator itOPCODEs_POST(post);
	
	for(; 
		!itOPCODEs_PRE.end() && !itOPCODEs_POST.end(); 
		++itOPCODEs_PRE, ++itOPCODEs_POST)
	{
		if (!OpcodesAreEqual(itOPCODEs_PRE, itOPCODEs_POST))
		{
			// Opcodes don't match
			return S_FALSE;
		}
		else if (itOPCODEs_PRE.isOpCodeEqual(CEE_LDSTR))
		{
			_ASSERTE(itOPCODEs_POST.isOpCodeEqual(CEE_LDSTR));

			// Compare two string references

			const DWORD* pdwStringAddrPre = (DWORD*) itOPCODEs_PRE.getOpCodeArg();
			const DWORD* pdwStringAddrPost = (DWORD*) itOPCODEs_POST.getOpCodeArg();

			_ASSERTE(pdwStringAddrPre && pdwStringAddrPost);
			if (!(pdwStringAddrPre && pdwStringAddrPost))
				return E_UNEXPECTED; // Return true on error so we ignore the method

			DWORD dwStringAddrPre = *pdwStringAddrPre;
			DWORD dwStringAddrPost = *pdwStringAddrPost;

			for(TypeListTokenPairs::const_iterator it = listEquivalentStringTokens.begin();
				it != listEquivalentStringTokens.end();
				++it)
			{
				const mdString& tokStringFrom = (*it).first;
				if (tokStringFrom == dwStringAddrPost)
				{
					// Ideally, we could patch up the method right here
					const mdString& tokStringTo = (*it).second;
					dwStringAddrPost = tokStringTo;
					break;
				}
			}

			if (dwStringAddrPre != dwStringAddrPost)
			{
				// String references don't refer to the same string value
				return S_FALSE;
			}
		}
		else if (!InstructionsAreEqual(itOPCODEs_PRE, itOPCODEs_POST))
		{
			InstructionsAreEqual(itOPCODEs_PRE, itOPCODEs_POST); // REMOVE ME!!! TODO - debugging only
			// Opcodes are the same but the opcode params are different
			return S_FALSE;
		}

	} // for each instruction

	if (itOPCODEs_PRE.isError() || itOPCODEs_POST.isError())
	{
		// There was an IL parsing error so ignore this method
		return E_FAIL;
	}
	else
	{
		// Returns true if we reached the end of both methods at the same time
		return itOPCODEs_PRE.end() && itOPCODEs_POST.end() ? S_OK : S_FALSE;
	}
}

HRESULT GetMethodNameFromFunctionId(	ICorProfilerInfo* pProfInfo,
										FunctionID functionId,
										wchar_t* wszClass, 
										wchar_t* wszMethod,
										DWORD dwSize)
{
	HRESULT hr = S_OK;

	CMethodInfo methodInfo;
	hr = methodInfo.init(pProfInfo, functionId, 0, 0, false);
    if (FAILED(hr)) return hr;

	wcsncpy(wszClass, methodInfo.getTypeName(), dwSize);
	wszClass[dwSize-1] = '\0';

	wcsncpy(wszMethod, methodInfo.m_szMemberName, dwSize);
	wszMethod[dwSize-1] = '\0';

    return S_OK;
}

HRESULT GetMethodPtrInfo(	LPCBYTE const pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							ULONG* pulMethodSize,
							unsigned __int8** ppbMethodHeader,
							ULONG* pulMethodHeaderSize,
							LPCBYTE* ppbMethodIL,
							ULONG* pulCodeSize,
							mdSignature* ptokLocalVarSig)
{
	if (!pbMapViewOfFile) return E_POINTER;
	if (!ppbMethodHeader && !ppbMethodIL && !pulCodeSize) return E_POINTER;

	HRESULT hr = S_OK;

	unsigned __int8* pbMethodHeader = 0;
	hr = GetMethodHeader(pbMapViewOfFile, rva, bIsImage, &pbMethodHeader);
	if (FAILED(hr)) return hr;
	if (!pbMethodHeader) return E_UNEXPECTED;

	if (ppbMethodHeader)
		*ppbMethodHeader = pbMethodHeader;

	return GetMethodIL(pbMethodHeader, pulMethodSize, pulMethodHeaderSize, ppbMethodIL,
		pulCodeSize, ptokLocalVarSig);
}

// Adapted from a presentation by John Lam
HRESULT GetMethodHeader(	LPCBYTE pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							unsigned __int8** ppbMethodHeader)
{
	HRESULT hr = S_OK;

	_ASSERTE(ppbMethodHeader);
	if (!ppbMethodHeader) return E_POINTER;

	if (bIsImage)
	{
		PIMAGE_NT_HEADERS pNtHeader = ::ImageNtHeader((PVOID) pbMapViewOfFile);
		if (!pNtHeader)
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());
			return ILXLOG(hr, _T("ImageNtHeader failed."), ILXLOG_ERROR);
		}

		*ppbMethodHeader = 
			(unsigned __int8*) ImageRvaToVa(pNtHeader, (PVOID) pbMapViewOfFile, rva, 0);
	}

	// KLUDGE: An RVA can either be a direct offset from the base address or
	// a virtual address (VA) relative to a section of a PE file.  It appears
	// to always be an offset if the CLR is loading the file, except when
	// running under Rotor where it always appears to be a VA.  It always
	// appears to be a VA when the base address comes from mapping a PE file
	// in memory.  So it appears that Rotor uses memory mapped files and the
	// CLR doesn't.  Possibly the CLR is obtaining the memory via IMethodAlloc
	// or something.  If ImageRvaToVa() doesn't appear to work we try treating
	// it as an offset.  This is very kludgy but I don't know another way to do this.

	if (!bIsImage) /* || IsBadMethodHeaderPtr(*ppbMethodHeader))*/
	{
		*ppbMethodHeader = (unsigned __int8*) pbMapViewOfFile + rva;
		//if (IsBadMethodHeaderPtr(*ppbMethodHeader))
		//{
		//	return ILXLOG(E_UNEXPECTED, _T("ImageRvaToVa() failed."), ILXLOG_ERROR);
		//}
	}

	return S_OK;
}

bool IsBadMethodHeaderPtr(unsigned __int8* pbMethodHeader)
{
	if (!pbMethodHeader)
	{
		return true;
	}

	COR_ILMETHOD* pMethodHeader = (COR_ILMETHOD*) pbMethodHeader;

	if (IsBadReadPtr(pMethodHeader, sizeof(COR_ILMETHOD)))
	{
		return true;
	}

	// Check for sane/legal values

	if (!pMethodHeader->Tiny.IsTiny() && !pMethodHeader->Fat.IsFat())
	{
		return true;
	}

	const DWORD MAX_CODE_SIZE_KLUDGE = 10000;
	if (pMethodHeader->Fat.IsFat() && pMethodHeader->Fat.CodeSize > MAX_CODE_SIZE_KLUDGE)
	{
		return true;
	}

	return false;
}

HRESULT GetMethodIL(	/*[in]*/  unsigned __int8* const pbMethodHeader,
						/*[out]*/ ULONG* pulMethodSize,
						/*[out]*/ ULONG* pulHeaderSize,
						/*[out]*/ LPCBYTE* ppbMethodIL,
						/*[out]*/ ULONG* pulCodeSize,
						/*[out]*/ mdSignature* ptokLocalVarSig)
{
	HRESULT hr = S_OK;

	_ASSERTE(pbMethodHeader);
	if (!pbMethodHeader) return E_POINTER;
	if (!ppbMethodIL && !pulHeaderSize && !pulCodeSize) return E_INVALIDARG;

	COR_ILMETHOD* pMethodHeader = 
		reinterpret_cast<COR_ILMETHOD*>(pbMethodHeader);

	COR_ILMETHOD_DECODER methodDecoder(pMethodHeader);

	if (ptokLocalVarSig)
	{
		*ptokLocalVarSig = methodDecoder.GetLocalVarSigTok() ? 
			methodDecoder.GetLocalVarSigTok() : mdSignatureNil;
	}

	if (ppbMethodIL)
		*ppbMethodIL = methodDecoder.Code;

	if (pulCodeSize)
		*pulCodeSize = methodDecoder.CodeSize;

	if (pulMethodSize)
		*pulMethodSize = methodDecoder.GetOnDiskSize(pMethodHeader);

	if (pulHeaderSize)
		*pulHeaderSize = (ULONG) (methodDecoder.Code - (BYTE*) pMethodHeader);

	return S_OK;
}
