///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MethodInfo.h 3     5/28/05 1:20p Eaddy $
//
//    File name:	MethodInfo.h
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MethodInfo.h $
 * 
 * 3     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/26 04:43:16  me133
 * no message
 *
 * Revision 1.1  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 */
#ifndef INCLUDED_METHOD_INFO_H
#define INCLUDED_METHOD_INFO_H

#include "MetadataIterators.h"	// mdMethodDef

#include "corhlpr.h"			// IMetaDataImport, mdMethodDef

#include <cor.h>				// Needed to compile corprof.h!
#include <corprof.h>			// ICorProfilerInfo

#include <list>					// STL list<>

// Type definitions

typedef std::pair<mdToken, mdToken>	TypePairTokens;
typedef std::list<TypePairTokens>	TypeListTokenPairs;

class CMethodInfo
{
public:

	CMethodInfo();
	CMethodInfo(const CMethodInfo& that);

	HRESULT init(methodDef_iterator& itMethodDef, const WCHAR* szTypeName, bool bIsImage);
	HRESULT init(ICorProfilerInfo* pProfInfo, FunctionID functionID, const WCHAR* szTypeName, LPCBYTE pbLoadAddress, bool bIsImage);
	HRESULT init(IMetaDataImport* piImport, mdMethodDef methodDef, const WCHAR* szTypeName, LPCBYTE pbLoadAddress, bool bIsImage);

	HRESULT updateMethod(	const CMethodInfo& methodNew,
							ModuleID moduleID,
							FunctionID functionID,
							IMethodMalloc* piAlloc);

	bool isPInvoke() const;
	bool isAbstract() const;

	LPBYTE getCodeStart() const;
	LPBYTE getCodeEnd() const;

	const WCHAR* getTypeName(); /* const */

private: // HELPER FUNCTIONS
	CMethodInfo& operator=(const CMethodInfo& that);

public: // DATA

	static const nMAX_ID = 2048; // from alink.h (Rotor)

	CComPtr<IMetaDataImport> m_spiImport;

	mdTypeDef m_typeDef;
	mdMethodDef m_methodDef;
    WCHAR m_szTypeName[nMAX_ID];
    WCHAR m_szMemberName[nMAX_ID];
    ULONG m_ulNameLen;
    DWORD m_dwFlags;
    PCCOR_SIGNATURE m_pbSigBlob;
    ULONG m_ulSigBlob;
    ULONG m_ulCodeRVA;
    ULONG m_ulImplFlags;

	unsigned __int8* m_pbMethodHeader;
	ULONG m_ulMethodSize;
	ULONG m_ulMethodHeaderSize;
	LPCBYTE m_pbMethodIL;
	ULONG m_ulCodeSize;

	mdSignature m_tokLocalVarSig;
	PCCOR_SIGNATURE m_pLocalVarSig;
	ULONG m_ulLocalVarSigSize;

	// Needed for method patching
	ModuleID m_moduleID;
	FunctionID m_functionID;
	bool m_bWasModified;
};

HRESULT GetMethodNameFromFunctionId(	ICorProfilerInfo* pProfInfo,
										FunctionID functionId,
										wchar_t* wszClass, 
										wchar_t* wszMethod,
										DWORD dwSize);

// Returns S_OK if the methods are equivalent, S_FALSE if the methods are
// not equivalent, and E_FAIL if a comparison couldn't be made
HRESULT MethodsAreEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenPairs& listNewOrReorderedStrings);

#endif // INCLUDED_METHOD_INFO_H
